/* All changes Copyright 2004, InfoGard Labs*/
/* Permission is granted to use this implementation for
 * non-commercial use
 */

#include <stdio.h>
#include <assert.h>
#include <openssl/des.h>
#include <openssl/opensslv.h>
#include "../include/config.h"

/*To support a compliant X9.31 A.2.4 RNG*/
void ANSIX931(BYTE *V, BYTE *DT, BYTE *R);
void ANSIX931_init(BYTE *Key);
void IncDTVector(BYTE *DT);
void Crypt931(BYTE* data, BYTE* result);
void xor_931(BYTE *A, BYTE *x, BYTE *y, int l);

/*ANSI X9.31 A.2.4*/

/* The original version of this was was not a correct implementation of X9.17.  
 * The PRNG implemented 
 * did not sample the DT value every round, and it regenerated the seed 
 * value incorrectly in an outer loop (the next V value was set to the
 * output by the generator in the ANSI function).
 * In addition, the code seem to be based off of a FIPS 186-2 framework, so
 * the code used 20 bytes for everything.  One out of three outputs from the
 * PRNG was truncated, decreasing the impact of the low order bits.
 * I have attempted to implement a compliant implementation 
 *
 * The variable names that I moved to reflect the X9.31 standard.
 * Appendix A.2.4:
 * Let ede*X(Y) represent the DEA multiple encryption of Y under the key
 * *X.  Let *K be a DEA key pair reserved only for the generation of
 * pseudo random numbers, let V be a 64-bit seed value which is also kept
 * secret, and let XOR be the exclusive-or operator.  Let DT be a
 * date/time vector which is updated on each iteration.  I is an
 * intermediate value.  A 64 bit vector R is generated as follows:
 *
 * I = ede*K(DT)
 * R = ede*K(I XOR V) and a new V is generated by V = ede*K(R XOR I).
 *
 * Successive values of R may be concatenated to produce a pseudo random
 * number of the desired length.
 * 
 * The underlying ANSIX9.31 PRNG passed the NIST CAVS tool on 15-Jul-2004
 * -- JEH 15-Jul-2004
 * 
 * After profiling the resulting code, I determined that using the openssl
 * TDES implementation provided me with roughly a 100x speedup, so this
 * version of the code uses the openssl crypto library.
 * Sadly, the OpenSSL group changed a number of the API names in 
 * openssl-0.9.7, which is the version that this code was written to.
 * If you're back in the "stone ages" and still use an OpenSSL version
 * that was release more than two weeks ago, you'll need to change
 * all the DES_ openssl api tags to des_.  In addition, there are some
 * basic variable changes (from an key array to a structure) that need
 * some changes.
 * -- JEH 14-Aug-2004
 *
 * I added some defines to deal with the various OpenSSL version
 * to make it a bit more reasonable
 * -- JEH 06-Nov-2004
 */
#ifndef OPENSSL_VERSION_NUMBER
#error OpenSSL version unknown
#endif

#if OPENSSL_VERSION_NUMBER < 0x00907000L
#define ANSIX931_ARRAY_DESKEY
#define ANSIX931_KEY des_key_schedule
#define ANSIX931_ENCRYPT des_ecb2_encrypt
#define ANSIX931_INBLOCK const_des_cblock
#define ANSIX931_OUTBLOCK des_cblock
#define ANSIX931_SETKEY des_set_key_unchecked
#else
#undef ANSIX931_ARRAY_DESKEY
#define ANSIX931_KEY DES_key_schedule
#define ANSIX931_ENCRYPT DES_ecb2_encrypt
#define ANSIX931_INBLOCK const_DES_cblock
#define ANSIX931_OUTBLOCK DES_cblock
#define ANSIX931_SETKEY DES_set_key_unchecked
#endif

ANSIX931_KEY key1, key2;

/**************************************************************************
* FUNCTION: ANSI X9.31 A.2.4 Pseudorandom Number Generator                *
*                                                                         *
* Generate random numbers according to the following:                     *
*	V0 = V;                                                           *
*	for j = 0 to (N-1) do {                                           *
*		I = DEAk(Tj)                                              *
*		R = DEAk(DEAk(I xor Vj) xor I)                            *
*		Vj+1 = R xor I                                            *
*       }                                                                 *
* Where V is a 64 bit seed, K is a 128 bit keypair, T represents a series *
* of distinct 64 bit date/time vectors (monotonically increasing)         *
*                                                                         *
* INPUT:                                       			          *
*    BYTE *V	- 64 bit seed					          *
*    BYTE *K 	- double key for encryption  (128 bit)                    *
* RETURNS:                                                                *
*    Random vectors in R space                                            *
*    Note: These comments were corrected. -- JEH
*    This implementation passed NIST CAVS testing on Jul 15
***************************************************************************/
void ANSIX931(BYTE *V, BYTE *DT, BYTE *R)
{
  BYTE I[8], IV[8];

  assert(sizeof(BYTE)==1);

  Crypt931(DT, I);	/* generate I */

  xor_931(IV, I, V, 8);
  Crypt931(IV, R);  	/* This is the output of the PRNG */
  xor_931(IV, I, R, 8);
  Crypt931(IV, V);/* Last encryption cycle */
  IncDTVector(DT); 	/* Increment vector */
}

/**************************************************************************
* FUNCTION: Increment Date/Time Vector                      	          *
*                                                                         *
* RETURNS:                                                                *
*    BYTE *dvector - 64 bit Date/Time vector		          	  *
***************************************************************************/
void IncDTVector(BYTE *DT)
{
  int cntr;
  BYTE *curPos;

  curPos = DT+7;
  for(cntr=0; cntr<8; cntr++) {
	  (*curPos)++;
	  if(*curPos!=0) {
		  break;
	  } else {
		  curPos--;
	  }
  }
}

void Crypt931(BYTE* data, BYTE* result)
/******************************************************************************
* Encrypts with a double length key 
*******************************************************************************/
{
  /* EDE */
#ifdef ANSIX931_ARRAY_DESKEY
  ANSIX931_ENCRYPT((ANSIX931_INBLOCK *)data, (ANSIX931_OUTBLOCK *)result, key1, key2, DES_ENCRYPT);
#else
  ANSIX931_ENCRYPT((ANSIX931_INBLOCK *)data, (ANSIX931_OUTBLOCK *)result, &key1, &key2, DES_ENCRYPT);
#endif
}

void ANSIX931_init(BYTE *Key) {
   assert(sizeof(BYTE) == 1);

#ifdef ANSIX931_ARRAY_DESKEY
   ANSIX931_SETKEY((ANSIX931_INBLOCK *)Key, key1);
   ANSIX931_SETKEY((ANSIX931_INBLOCK *)(Key+8), key2);
#else
   ANSIX931_SETKEY((ANSIX931_INBLOCK *)Key, &key1);
   ANSIX931_SETKEY((ANSIX931_INBLOCK *)(Key+8), &key2);
#endif
}

void xor_931(BYTE *A, BYTE *x, BYTE *y, int l)
{
  while (l--)  A[l] = x[l] ^ y[l];
}

