/* All changes Copyright 2004, InfoGard Labs*/
/* Permission is granted to use this implementation for
 * non-commercial use
 */

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <openssl/sha.h>
#include <openssl/des.h>
#include <openssl/opensslv.h>
#include "../include/config.h"

#ifndef OPENSSL_VERSION_NUMBER
#error OpenSSL version unknown
#endif

#if OPENSSL_VERSION_NUMBER < 0x00907000L
#define ANSIX931_ARRAY_DESKEY
#define FIPS186_KEY des_key_schedule
#define FIPS186_ENCRYPT des_ecb_encrypt
#define FIPS186_INBLOCK const_des_cblock
#define FIPS186_OUTBLOCK des_cblock
#define FIPS186_SETKEY des_set_key_unchecked
#else
#undef ANSIX931_ARRAY_DESKEY
#define FIPS186_KEY DES_key_schedule
#define FIPS186_ENCRYPT DES_ecb_encrypt
#define FIPS186_INBLOCK const_DES_cblock
#define FIPS186_OUTBLOCK DES_cblock
#define FIPS186_SETKEY DES_set_key_unchecked
#endif


/*The original code provided in the sts package was not a compliant 
 *version of FIPS 186-2.
 *The DES-G based generator had the following issues:
 *The generator initializes and acts on xval, and then doesn't use it.
 *The generator used XKEY instead of the specified 't' value.
 *The XSEED is used instead of the xval variable.
 *The DES_G function does not form the DES key correctly
 *The generator integrated Mod q into the feedback and not the output of 
 *the generator.
 *
 *The SHA-G based generator had the following issues:
 *It didn't hard code the IV (the IV is specified in the standard).
 *The provided IV was not compliant.
 *xval is not being updated each round correctly (it did not include
 *an input from the existing XKEY)
 */

/*To support a compliant FIPS 186-2 Ap. 3.1 general purpose RNG*/
/*The actual generator*/
void FIPS186_31(BYTE *xkey, BYTE *xseed, int xkeylen, BYTE *x, void (*G)(BYTE *, UINT32 *, BYTE *, int));

/*Support functions*/
/*This function outputs (x+xkey+1) mod 2^(8*len)*/
inline void fips186_add1(BYTE *out, BYTE *x, BYTE *xkey, int xkeylen);
/*This function outputs (xseed+xkey) mod 2^(8*len)*/
inline void fips186_add(BYTE *out, BYTE *xseed, BYTE *xkey, int xkeylen);

/*The G functions*/
void DES_G(BYTE *out, UINT32 *t, BYTE *c, int cLen);
void SHA_G(BYTE *out, UINT32 *t, BYTE *c, int cLen);
UINT32 DES_G_expandKey(UINT32 b1);
void DES_G_DES(UINT32 b1, UINT32 b2, UINT32 a1, UINT32 a2, BYTE *CT);
inline void byteToU32(BYTE *in, UINT32 *out);
inline void U32ToByte(UINT32 in, BYTE *out);
inline UINT32 FIPS186_betoh(UINT32 in);
inline UINT32 FIPS186_htobe(UINT32 in);
inline UINT32 FIPS186_byteswap(UINT32 in);

/*FIPS 186-2 Ap. 3.1 general purpose RNG*/

/*This function outputs (x+xkey+1) mod 2^(8*len)*/
/* add1 is only useful in the context of the FIPS 186-2 RNG
 * The code assumes that x is 160 bits (20 bytes)
 * The code assumes that xkey and out are xkeylen.
 */
inline void fips186_add1(BYTE *out, BYTE *x, BYTE *xkey, int xkeylen)
{
	unsigned int cntr, carry, value;

	for(cntr=1; cntr<=20; cntr++) {
		if(cntr==1) {
			value=xkey[xkeylen-1]+x[19]+1;
			out[xkeylen-1]=value;
			carry = (value - out[xkeylen-1])>>8;
		} else {
			value=xkey[xkeylen-cntr]+x[20-cntr]+carry;
			out[xkeylen-cntr]=value;
			carry=(value - out[xkeylen-cntr])>>8;
		}
	}

	for(cntr=21; cntr<=xkeylen; cntr++) {
		value=xkey[xkeylen-cntr]+carry;
		out[xkeylen-cntr]=value;
		carry = (value-out[xkeylen-cntr])>>8;
	}
}

/*This function outputs (xseed+xkey) mod 2^(8*len)*/
/* add is only useful in the context of the FIPS 186-2 RNG
 * The code assumes that xseed, xkey and out are xkeylen.
*/
inline void fips186_add(BYTE *out, BYTE *xseed, BYTE *xkey, int xkeylen)
{
	unsigned int cntr, value, carry=0;

	for(cntr=1; cntr<=xkeylen; cntr++) {
		value=xkey[xkeylen-cntr]+xseed[xkeylen-cntr]+carry;
		out[xkeylen-cntr]=value;
		carry = (value-out[xkeylen-cntr])>>8;
	}
}

/*The G functions*/
/*The DES_G function only works is cLen == 20*/
/* This is the DES based G function.  It would appear that
 * This was intended for hardware implementation, given the
 * obnoxious bit-oriented manipulation that is required.
 * 
 * Note: This was implemented to be as obviously consistent with
 * the standard as possible, with one proviso: X uses 0 indexed arrays,
 * so the indexing is slightly different, which lead to a slight 
 * modification of the formulas used to calculate the active index
 * in each case.  A few quick calculations should show that they are
 * equivalent.
 * 
 * The step references are from FIPS 186-2 Ap 3.4.
 */
void DES_G(BYTE *out, UINT32 *t, BYTE *ci, int cLen)
{
	UINT32 c[5], x[5], y[5][2], z[5];
	UINT32 b1, b2, a1, a2;
	BYTE CT[8];
	int i;

	if(cLen != 20) {
		fprintf(stderr, "DES based G does not support xkeylen other than 20\n");
		exit(1);
	}

	/*Step 1*/
	/*divide up c[i]*/
	byteToU32(ci, &c[0]);
	byteToU32(ci+4, &c[1]);
	byteToU32(ci+8, &c[2]);
	byteToU32(ci+12, &c[3]);
	byteToU32(ci+16, &c[4]);

	/*Step 2*/
	/*Populate x[i]*/
	for(i=0; i<5; i++)
		x[i] = t[i] ^ c[i];

	/*Step 3*/
	/*Fill in the y matrix*/
	for(i=0; i<5; i++) {
		b1 = c[(i+4)%5];
		b2 = c[(i+3)%5];
		a1 = x[i];
		a2 = x[(i+1)%5] ^ x[(i+4)%5];
		DES_G_DES(b1, b2, a1, a2, CT);
		byteToU32(CT, &y[i][0]);
		byteToU32(CT+4, &(y[i][1]));
	}

	/*Step 4*/
	/*Populate z[i]*/
	for(i=0; i<5; i++)
		z[i] = y[i][0] ^ y[(i+2)%5][1] ^ y[(i+3)%5][0];

	/*Step 5*/
	/*return the result*/
	U32ToByte(z[0], out);
	U32ToByte(z[1], out+4);
	U32ToByte(z[2], out+8);
	U32ToByte(z[3], out+12);
	U32ToByte(z[4], out+16);
}

/*This function performs the DES variant required by the standard.
 *Note that there are some defines used here for OpenSSL version
 *differences.
 */

void DES_G_DES(UINT32 b1, UINT32 b2, UINT32 a1, UINT32 a2, BYTE *CT)
{
	UINT32 tb1, tb2;
	BYTE key[8], PT[8];
	FIPS186_KEY DES_G_sched;
	/*Create the key*/

	/*convert these to local endianess for bitwise operations*/
	tb1 = FIPS186_betoh(b1);
	tb2 = FIPS186_betoh(b2);

	/*We want to move the highest order nibble of b2 (which is completely
         *present in the resulting key) to into the low order nibble of b1 
	 *(only the low order 24 bits of this key are present in the 
         *resulting key)*/
	tb1 = tb1 << 4;

	tb2 &= 0xF0000000;
	tb2 = tb2 >> 28;

	tb1 |= tb2;

	tb2 = FIPS186_betoh(b2);

	/*We want to discard the high order nibble of b2, but this happens
	 *automatically in the DES_G_expandKey function*/
	tb1 = DES_G_expandKey(tb1);
	tb2 = DES_G_expandKey(tb2);

	/* Back to big endian for DES */
	b1 = FIPS186_htobe(tb1);
	b2 = FIPS186_htobe(tb2);

	U32ToByte(b1, key);
	U32ToByte(b2, key+4);
#ifdef ANSIX931_ARRAY_DESKEY
	FIPS186_SETKEY((FIPS186_INBLOCK *)key, DES_G_sched);
#else
	FIPS186_SETKEY((FIPS186_INBLOCK *)key, &DES_G_sched);
#endif

	/*now populate the plaintext*/
	U32ToByte(a1, PT);
	U32ToByte(a2, PT+4);

	/*turn the DES crank*/
#ifdef ANSIX931_ARRAY_DESKEY
	FIPS186_ENCRYPT((FIPS186_INBLOCK *)PT, (FIPS186_OUTBLOCK *)CT, DES_G_sched, DES_ENCRYPT);
#else
	FIPS186_ENCRYPT((FIPS186_INBLOCK *)PT, (FIPS186_OUTBLOCK *)CT, &DES_G_sched, DES_ENCRYPT);
#endif

}

/*We are provided with an unsigned int with data in the lower 7 nibbles *
 *We need to spread this data among the bytes, leaving the parity bit  *
 *out (The parity bit for DES is the lowest order bit of each byte.)    *
 * Spread happens from low order to high order			        *
 * This code accomplishes last last part of the translation from 
 * (b1,b2) to the key used by the DES engine.
 * This function requires that the integers fed to it are in machine
 * local endianess. (so the shifts and masking works correctly)
 */
UINT32 DES_G_expandKey(UINT32 b1)
{
	UINT32 k1, k2, k3, k4;

	/*shift up the bits for the most significant byte*/
	k1 = b1 << 4;
	/*Discard the low order bit in the byte (DES Parity)*/
	k1 &= 0xFE000000;

	/*shift up the bits for the next byte*/
	k2 = b1 << 3;
	k2 &= 0x00FE0000;


	/*shift up the bits for the next byte*/
	k3 = b1 << 2;
	k3 &= 0x0000FE00;


	/*shift up the bits for the last byte*/
	k4 = b1 << 1;
	k4 &= 0x000000FE;

	return (k1|k2|k3|k4);
}

/*The SHA-1-ish based G function.  It's prettier than the
 *DES based one, but it requires a silly amount of SHA-1
 *implementation magic.  This code relies on the fact that the
 *OpenSSL SHA-1 implementation keeps the hash value updated
 *per block, and that we can both read and write to a container
 *that ought to be opaque.  These assumptions could break as they
 *are not guaranteed by the library (These are _not_ 
 *published APIs, and even if they were, with openssl, they
 *probably wouldn't be stable, anyway...)
 *
 * This function takes in a 20 byte 't' value (established by
 * the generator used), and a 20-64 byte 'c' value.
 * It pads the block as required, and then performs SHA-1 steps
 * a-e on the data.
 */
void SHA_G(BYTE *out, UINT32 *t, BYTE *c, int cLen)
{
	BYTE block[64];
	SHA_CTX shaContext;
	UINT32 tval[5];
	int cntr;

	for(cntr=0; cntr<5; cntr++) {
		tval[cntr]= FIPS186_betoh(t[cntr]);
	}

	/*Copy the block for padding*/
	memcpy(block, c, cLen);

	/*Pad the block if necessary*/
	if(cLen < 64)
		memset(&block[cLen], 0, 64-cLen);

	/*Magic to go steps a-e of SHA-1*/
	SHA1_Init(&shaContext);
	shaContext.h0=tval[0];
	shaContext.h1=tval[1];
	shaContext.h2=tval[2];
	shaContext.h3=tval[3];
	shaContext.h4=tval[4];
	SHA1_Update(&shaContext, block, 64);
	U32ToByte(FIPS186_htobe(shaContext.h0), out);
	U32ToByte(FIPS186_htobe(shaContext.h1), out+4);
	U32ToByte(FIPS186_htobe(shaContext.h2), out+8);
	U32ToByte(FIPS186_htobe(shaContext.h3), out+12);
	U32ToByte(FIPS186_htobe(shaContext.h4), out+16);
	/*end magic*/
}

/*The Actual generator*/
/*
 * xkey must be xkeylen bytes long
 * xseed must be either null (if no input) or xkeylen bytes long
 * 20 <= xkeylen <= 64 (if DES_G is used, it must be 20)
 * x must be 20 bytes long (160 bits)
 * G should be either DES_G or SHA_G
*/
void FIPS186_31(BYTE *xkey, BYTE *xseed, int xkeylen, BYTE *x, void (*G)(BYTE *, UINT32 *, BYTE *, int))
{
	BYTE xval[64]; /*Can't be larger than 64...*/
	BYTE newxkey[64];
	UINT32 tx[5] = {0x67452301, 0xefcdab89, 0x98badcfe, 0x10325476, 0xc3d2e1f0};
	UINT32 t[5];
	int i;

	assert(sizeof(BYTE) == 1);
	assert(sizeof(UINT32) == 4);

	if((xkeylen < 20) || (xkeylen > 64)) {
		fprintf(stderr, "20 <= xkeylen <= 64\n");
		exit(1);
	}

	for(i=0; i<5; i++)
		t[i] = FIPS186_htobe(tx[i]);

	if(xseed)
		fips186_add(xval, xseed, xkey, xkeylen);
	else
		memcpy(xval, xkey, xkeylen);

	G(x, t, xval, xkeylen);

	fips186_add1(newxkey, x, xkey, xkeylen);
	memcpy(xkey, newxkey, xkeylen);
}

inline void byteToU32(BYTE *in, UINT32 *out)
{
	memcpy(out, in, 4);
}

inline void U32ToByte(UINT32 in, BYTE *out)
{
	memcpy(out, &in, 4);
}

inline UINT32 FIPS186_htobe(UINT32 in)
{
#ifdef STS_LITTLE_ENDIAN
	return FIPS186_byteswap(in);
#else
	return in;
#endif
}

inline UINT32 FIPS186_betoh(UINT32 in)
{
#ifdef STS_LITTLE_ENDIAN
	return FIPS186_byteswap(in);
#else
	return in;
#endif
}

inline UINT32 FIPS186_byteswap(UINT32 in)
{
	return (((in>>24)&0x000000FF)|((in>>8)&0x0000FF00)|((in<<8)&0x00FF0000)|((in<<24)&0xFF000000));
}
