/* --------------------------------------------------------------------------
   Title       :  The NIST Statistical Test Suite

   Date        :  December 1999

   Programmer  :  Juan Soto

   Summary     :  For use in the evaluation of the randomness of bitstreams
		  produced by cryptographic random number generators.

   Package     :  Version 1.0

   Copyright   :  (c) 1999 by the National Institute Of Standards & Technology

   History     :  Version 1.0 by J. Soto, October 1999
		  Revised by J. Soto, November 1999

   Keywords    :  Pseudorandom Number Generator (PRNG), Randomness, Statistical 
                  Tests, Complementary Error functions, Incomplete Gamma 
	          Function, Random Walks, Rank, Fast Fourier Transform, 
                  Template, Cryptographically Secure PRNG (CSPRNG),
		  Approximate Entropy (ApEn), Secure Hash Algorithm (SHA-1), 
                  Blum-Blum-Shub (BBS) CSPRNG, Micali-Schnorr (MS) CSPRNG, 

   Source      :  David Banks, Elaine Barker, James Dray, Allen Heckert, 
		  Stefan Leigh, Mark Levenson, James Nechvatal, Andrew Rukhin, 
		  Miles Smid, Juan Soto, Mark Vangel, and San Vo.

   Technical
   Assistance  :  Lawrence Bassham, Ron Boisvert, James Filliben, Sharon Keller,
		  Daniel Lozier, and Bert Rust.

   Warning     :  Portability Issues.

   Limitation  :  Amount of memory allocated for workspace.

   Restrictions:  Permission to use, copy, and modify this software without 
		  fee is hereby granted, provided that this entire notice is 
		  included in all copies of any software which is or includes
                  a copy or modification of this software and in all copies 
                  of the supporting documentation for such software.
   -------------------------------------------------------------------------- */

/* 
 * Made the "passrate" a range rather than a minimum value and added acceptable 
 * "passrate" checking for computeMetrics rather than the hard-coded 0.96 rate. 
 * -- JEH 2004-Jul-16
 *
 * Made the ANSI generator refer to the new (checked) ANSI-X9.31 RNG rather 
 * than the (non-compliant) ANSI 9.17 RNG.
 * The X9.31 and X9.17 generators are equivalent if implemented correctly.
 * -- JEH 2004-Jul-14
 *
 * Minor cleanup. Corrected some minor C syntax issues, 
 * removed unused variables.
 * 
 * Added support for parameter validation (as per SP800-22).
 *
 * Modified the reporting routine for Lempel-Ziv in order to integrate
 * the suggested weighting proposed by KUH.
 *
 * Added a bit more data to the finalAnalysisReport.  The report now 
 * includes a list of the settings for the testing, an assessment
 * of the appropriateness of the settings, and the total amount of time
 * spent testing.
 * -- JEH 2004-Aug-15
 *
 * Some minor cleanup to banish compiler warnings.
 * -- JEH 2004-Oct-01
 *
 * Centralized global declaration here.  This change triggered a series
 * of minor include changes in several of the source files.
 *  
 * Moved post, grid and output opening to openOutputStreams in order
 * to better support the new configuration settings.
 *
 * Changed initial startup logic ordering to support the new alternate
 * command line.
 *
 * Added initialization for the new configuration global variables,
 * and performed initialization on startup to allow for conditional
 * prompts if the variable is not defined.
 *
 * Added some high level comments.
 *
 * Made the algorithm changes suggested by KUH optional
 *
 * Made the KUH variants flagged in the final report.
 * -- JEH 2004-Oct-08
 *
 * Added the option of a reporting-only style processing (for supporting
 * very long runs)
 * -- JEH 2004-Oct-10
 *
 * Added a number of asserts to catch odd conditions for all the tests.
 * Also verified that all memory allocations are checked for failure.
 * Made assess output the parameters of the current test to stdout.
 * 
 * Changed the "flag" value that indicates that there are not a sufficient
 * number of excursions (which was 0.0) to something outside the expected
 * range of the P value (-1.0).  This should make it so that extreem failures
 * don't look like the test just failing to work.
 * -- JEH 2004-Oct-17
 *
 * Removed depricated headers (malloc.h).
 * -- JEH 2004-Nov-11
 *
 * A few changes to prevent array overruns, and to prevent memory leaks
 * -- JEH 2007-Aug-12
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <assert.h>
#include "../include/defs.h"
#include "../include/proto.h"
#include "../include/special-functions.h"
#include "../include/mconf.h"  
#include "../include/utilities1.h"
#include "../include/utilities2.h"

/*Globals are defined once here, and references as externs elsewhere...*/

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
                   G L O B A L   D A T A  S T R U C T U R E S 
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

BitField*   epsilon;			/* BIT STREAM 	            */
BitField*   template;			/* TEMPLATE 	            */
TP	    tp;				/* TEST PARAMETER STRUCTURE */
CO          config;                     /* GENERATOR SEED AND CONFIG STRUCTURE */
FILE*       stats[NUMOFTESTS+1];	/* FILE OUTPUT STREAM       */
FILE*       results[NUMOFTESTS+1];	/* FILE OUTPUT STREAM       */
FILE*       output;			/* FILE OUTPUT STREAM       */
FILE*       post;			/* FILE OUTPUT STREAM       */
FILE*       cycleInfo;			/* FILE OUTPUT STREAM       */
FILE*       grid;			/* FILE OUTPUT STREAM       */
int         testVector[NUMOFTESTS+1];

void postProcessResults(int);
int cmp(const void* a, const void* b);
int computeMetrics(char*,int);

int main(int argc, char *argv[])
{
  int    i;
  int    option;		/* TEMPLATE LENGTH/STREAM LENGTH/GENERATOR*/
  int    commandLineError = 0;
  char*  streamFile;		/* STREAM FILENAME     */
  time_t startTime=0, endTime=0;

   /*Initialize all the variables to default values*/
   tp.n = -1;
   tp.blockFrequencyBlockLength = -1;
   tp.nonOverlappingTemplateBlockLength = -1;
   tp.overlappingTemplateBlockLength = -1;
   tp.universalBlockLength = -1;
   tp.universalNumberInitializationSteps = -1;
   tp.approximateEntropyBlockLength = -1;
   tp.serialBlockLength = -1;
   tp.linearComplexitySequenceLength = -1;
   tp.numOfBitStreams = -1;
   tp.validParams = -1;
   tp.useCorrectedLZ = 1;

   config.baseDir = NULL;
   config.configFile = NULL;
   config.generatorOption = -1;
   config.fips186.xkey = NULL;
   config.ansix931.K = NULL;
   config.ansix931.DT = NULL;
   config.ansix931.V = NULL;
   config.dataFile = NULL;
   config.dataFileFormat = NO_FORMAT;
   config.testsSelected = 0;

   for(i=0; i<NUMOFTESTS+1; i++) stats[i]=NULL;

  /*Deal with a the command line */
  if(argc == 2) {
     tp.n = atoi(argv[1]);
  } else if (argc == 3) {
     if(!strcmp(argv[1], "-c"))
        processConfigFile(argv[2]);
     else
        commandLineError = 1;
  } else
     commandLineError = 1;

  if(commandLineError) {
     printf("-----------------------------------------------------------\n");
     printf("        P A R A M E T E R       R E P R E S E N T A T I O N\n");
     printf("       --------------------    ----------------------------\n");
     printf("Usage: STREAM-SIZE          => BITSTREAM LENGTH.           \n");
     printf("                        or                                 \n");
     printf("       -c <config file>     => Read in a configuration file\n");
     printf("-----------------------------------------------------------\n");
     exit(1);
  }

   /*First grab any information that we don't have*/ 
   option = generatorOptions(&streamFile);
   chooseTests();
   fixParameters();
   validateParams();
   printParams(stdout);

   /*Now open all the files*/
   openOutputStreams(option);

   if(config.generatorOption != 12) { /*Should we actually do testing?*/

      /*Start the clock, run the tests, stop the clock*/
      startTime = time(NULL);
      invokeTestSuite(option,streamFile);
      endTime = time(NULL);

      /*close out the running status files*/
      fclose(output);
      fclose(grid);

      for(i = 0; i < NUMOFTESTS+1; i++) {
         if (stats[i] != NULL) fclose(stats[i]);
         if (results[i] != NULL) fclose(results[i]);
      }

   }

   /*Make the reports*/
   if (testVector[0] == 1 || testVector[3] == 1) 
      partitionResultFile(2, tp.numOfBitStreams, option, 0);
   if (testVector[0] == 1 || testVector[8] == 1) 
      partitionResultFile(MAXNUMOFTEMPLATES, tp.numOfBitStreams, option,1);
   if (testVector[0] == 1 || testVector[12] == 1)
      partitionResultFile(8, tp.numOfBitStreams, option, 2);
   if (testVector[0] == 1 || testVector[13] == 1)
      partitionResultFile(18, tp.numOfBitStreams, option, 3);
   if (testVector[0] == 1 || testVector[14] == 1)
      partitionResultFile(2, tp.numOfBitStreams, option, 4);
   printParams(post);
   fprintf(post,"Total time testing: %ld seconds\n", (long int) endTime-startTime);
   fprintf(post,"-----------------------------------------------");
   fprintf(post,"-------------------------------\n");
   fprintf(post,"RESULTS FOR THE UNIFORMITY OF P-VALUES AND THE ");
   fprintf(post,"PROPORTION OF PASSING SEQUENCES\n");
   fprintf(post,"-----------------------------------------------");
   fprintf(post,"-------------------------------\n\n");
   fprintf(post,"-----------------------------------------------");
   fprintf(post,"-------------------------------\n");
   fprintf(post," C1  C2  C3  C4  C5  C6  C7  C8  C9 C10");
   fprintf(post,"  P-VALUE  PROPORTION  STATISTICAL TEST\n");
   fprintf(post,"-----------------------------------------------");
   fprintf(post,"-------------------------------\n");
   postProcessResults(option);
   fclose(post);
   if(config.baseDir) free(config.baseDir);
   if(config.dataFile) free(config.dataFile);
   if(config.ansix931.V) free(config.ansix931.V);
   if(config.ansix931.DT) free(config.ansix931.DT);
   if(config.ansix931.K) free(config.ansix931.K);
   if(config.fips186.xkey) free(config.fips186.xkey);
   return 1;
}

void partitionResultFile(int numOfFiles, int numOfSequences, int option, 
                         int category)
{
  int    i, k, m, j, start, end, num, numread;
  float  c;
  FILE** fp = (FILE**)calloc(numOfFiles+1, sizeof(FILE*));
  char   **s = (char **)calloc(numOfFiles, sizeof(char *));
  char   resultsDir[200];
  char   *baseDir;
  char   generatorDir[NUMOFGENERATORS][20] = {"AlgorithmTesting","G-SHA-1",
         "LCG", "BBS", "MS", "MODEXP", "QCG1", "QCG2","CCG", "XOR", "ANSI-X9.31", 
         "G-DES"};
  char   testNames[5][50] = {"cumulative-sums","nonperiodic-templates",
			    "random-excursions", "random-excursions-variant",
			    "serial"};

  if(!fp) {
     printf("Out of memory for fp in partitionResultFile\n");
     exit(-1);
  }
  if(!s) {
     printf("Out of memory for s in partitionResultFile\n");
     exit(-1);
  }

  for(i=0; i < numOfFiles; i++) {
     s[i] = (char*)calloc(200,sizeof(char));
     if(!s[i]) {
       printf("Out of memory for s in partitionResultFile\n");
       exit(-1);
     }
  }

  if(config.baseDir)
     baseDir = config.baseDir;
  else
     baseDir = generatorDir[option];

  sprintf(resultsDir,"experiments/%s/%s/results", baseDir,
          testNames[category]);
  if (DISPLAY_OUTPUT_CHANNELS) printf("OPEN FILE:  %s\n", resultsDir);

  if ( (fp[numOfFiles] = fopen(resultsDir,"r")) == NULL ) {
     printf("%s",resultsDir);
     printf(" -- file not found. Exiting program.\n");
     exit(-1);
  }

  for(i = 0; i < numOfFiles; i++) {
     if (i < 10)
        sprintf(s[i],"experiments/%s/%s/data%1d", baseDir,
                testNames[category],i+1);
     else if (i < 100)
        sprintf(s[i],"experiments/%s/%s/data%2d", baseDir,
                testNames[category],i+1);
     else
        sprintf(s[i],"experiments/%s/%s/data%3d", baseDir,
                testNames[category],i+1);
     if (DISPLAY_OUTPUT_CHANNELS) printf("OPEN = %s\n", s[i]);
  }
  numread = 0;
  m = numOfFiles/20;
  if ((numOfFiles%20) != 0) m++;
  for(i = 0; i < numOfFiles; i++) {
     if ( (fp[i] = fopen(s[i],"w")) == NULL ) {
		 printf("%s",s[i]);
		 printf(" -- file not found. Exiting program.\n");
		 exit(-1);
     }
     fclose(fp[i]);
  }
  for(num = 0; num < numOfSequences; num++) {
     for(k = 0; k < m; k++) { 			/* FOR EACH BATCH */

       start = k*20;		/* BOUNDARY SEGMENTS */
       end   = k*20+19;
       if ( k == (m-1) )
           end = numOfFiles-1;

        if (PARTITION) printf("k = %d start = %d end = %d\n", k, start, end);

        for(i = start; i <= end; i++) {		/* OPEN FILE */
           if ( (fp[i] = fopen(s[i],"a")) == NULL ) {
			   printf("%s",s[i]);
			   printf(" -- file not found. Exiting program.\n");
			   exit(-1);
			   }
           if (PARTITION) printf("Open file %s\n", s[i]);
		   }

        for(j = start; j <= end; j++) {		/* POPULATE FILE */
            if(fscanf(fp[numOfFiles],"%f", &c) == EOF) {
              printf("Unexpected EOF populating parition file\n");
			     exit(-1);
            }

            fprintf(fp[j], "%f\n", c);
            numread++;
            if (PARTITION) printf("NUMREAD = %d\n",numread);
        }

        for(i = start; i <= end; i++) {		/* CLOSE FILE */
           fclose(fp[i]);
           if (PARTITION) printf("Close file %s\n", s[i]);
        }
     }
  }
  fclose(fp[numOfFiles]);
  for(i=0; i < numOfFiles; i++) free(s[i]);
  free(s);
  free(fp);
  return;
}

int cmp(const void* a, const void* b)
{
  if (*(const double*)a < *(const double*)b) return -1;
  if (*(const double*)a == *(const double*)b) return 0;
  return 1;
}

void postProcessResults(int option)
{
  int    i, k, randomExcursionSampleSize, generalSampleSize;
  int    case1, case2, numOfFiles = 2;
  double passRateMin, passRateMax;
  char   s[200];
  char   *baseDir;
  char   generatorDir[NUMOFGENERATORS][20] = {"AlgorithmTesting","G-SHA-1",
         "LCG", "BBS", "MS", "MODEXP", "QCG1", "QCG2","CCG", "XOR", "ANSI-X9.31", 
         "G-DES"};
  char   testNames[NUMOFTESTS+1][50] = {" ", "frequency", "block-frequency",
         "cumulative-sums", "runs", "longest-run", "rank", "fft",
         "nonperiodic-templates", "overlapping-templates", "universal", "apen",
         "random-excursions", "random-excursions-variant", "serial",
         "lempel-ziv", "linear-complexity"}; /* , "opso"; */

  assert(post);

  if(config.baseDir)
     baseDir = config.baseDir;
  else
     baseDir = generatorDir[option];

  for(i = 1; i <= NUMOFTESTS; i++) {		/* FOR EACH TEST */

     if (testVector[i] == 1) {

	/* SPECIAL CASES -- HANDLING MULTIPLE FILES FOR A SINGLE TEST */

        if ((i ==  3 && testVector[3] ) || (i ==  8 && testVector[8] ) ||
            (i == 12 && testVector[12]) || (i == 13 && testVector[13]) || 
            (i == 14 && testVector[14])) {

           if (i == 8 && testVector[8])  
              numOfFiles = MAXNUMOFTEMPLATES;
           else if (i == 12 && testVector[12]) 
              numOfFiles = 8;
           else if (i == 13 && testVector[13]) 
              numOfFiles = 18;
           else
              numOfFiles = 2;
           for(k = 0; k < numOfFiles; k++) {
              if (k < 10)
                 sprintf(s,"experiments/%s/%s/data%1d", baseDir,
                         testNames[i],k+1);
              else if (k < 100)
                 sprintf(s,"experiments/%s/%s/data%2d", baseDir,
                         testNames[i],k+1);
              else
                 sprintf(s,"experiments/%s/%s/data%3d", baseDir,
                         testNames[i],k+1);
 	      generalSampleSize = computeMetrics(s,i);
	      if (i == 12 || i == 13) 
                 randomExcursionSampleSize = generalSampleSize;
           }
        }
	else {
           sprintf(s,"experiments/%s/%s/results", baseDir,
                   testNames[i]);
	   generalSampleSize = computeMetrics(s,i);
	}
     }
  }

  fprintf(post, "\n\n- - - - - - - - - - - - - - - - - - - - - - - - - - - -");
  fprintf(post, "- - - - - - - - - - - - -\n");
  case1 = 0;
  case2 = 0;
  if (testVector[12] || testVector[13]) 
     case2 = 1;
  for(i = 1; i <= NUMOFTESTS; i++) {
     if (testVector[i] && i != 12 && i != 13) {
        case1 = 1;
        break;
     }
  }
  if (case1) {
     passRateMin = 0.99-3.0*sqrt(0.01*(1.0-ALPHA)/(double)generalSampleSize);
     passRateMax = 0.99+3.0*sqrt(0.01*(1.0-ALPHA)/(double)generalSampleSize);
     fprintf(post, "The minimum pass rate for each statistical test with");
     fprintf(post, "the exception of the\nrandom excursion (variant) test ");
     fprintf(post, "is approximately = %f \n",passRateMin);
     fprintf(post, "the maximum pass rate for each statistical test with");
     fprintf(post, "the exception of the\nrandom excursion (variant) test ");
     fprintf(post, "is approximately = %f\n",(passRateMax > 1.0)?1.0:passRateMax);
     fprintf(post, "for a sample size = %d binary sequences.\n\n", generalSampleSize);
  }
  if (case2) {
     passRateMin =.99-3.*sqrt(.01*(1.-ALPHA)/(double)randomExcursionSampleSize);
     passRateMax =.99+3.*sqrt(.01*(1.-ALPHA)/(double)randomExcursionSampleSize);
     fprintf(post, "The minimum pass rate for the random excursion");
     fprintf(post, "(variant) test is approximately = %f\n", passRateMin);
     fprintf(post, "The maximum pass rate for the random excursion");
     fprintf(post, "(variant) test is approximately = %f\n", (passRateMax > 1.0)?1.0:passRateMax);
     fprintf(post, "for a sample size = %d binary sequences.\n\n", 
             randomExcursionSampleSize);
  }
  fprintf(post, "For further guidelines construct a probability table using\n");
  fprintf(post, "the MAPLE program provided in the addendum section of the\n");
  fprintf(post, "documentation.\n");
  fprintf(post, "- - - - - - - - - - - - - - - - - - - - - - - - - - - -");
  fprintf(post, "- - - - - - - - - - - - -\n");
  return;
}

int computeMetrics(char* s, int test)
{
  int    j, pos, count, sampleSize, expCount;
  int    freqPerBin[10] = {0,0,0,0,0,0,0,0,0,0};
  double *A, *T, chi2, proportion, uniformity, passRateMin, passRateMax;
  float  c;
  FILE*  fp;
  char   testNames[NUMOFTESTS+1][50] = {" ", "Frequency", "Block-Frequency",
         "Cusum", "Runs", "Long-Run", "Rank", "FFT", "Aperiodic-Template", 
         "Periodic-Template", "Universal", "Apen", "Random-Excursion", 
         "Random-Excursion-V", "Serial", "Lempel-Ziv", "Linear-Complexity"}; 
  /* The fudge factor for the goodness-of-fit test for the Lempel-Ziv test
   * as suggested by Kim, Umeno and Hasegawa */
  double LZ_S[10] = { 0.1097085, 0.079127, 0.107691, 0.084465, 0.1369235,
	0.091115, 0.0858035, 0.1098615, 0.1028565, 0.0924485 };

  assert(s);
  assert(post);

  if (DISPLAY_OUTPUT_CHANNELS) printf("OPEN = %s\n", s);
  if ( (fp = fopen(s,"r")) == NULL ) {
     printf("%s",s);
     printf(" -- file not found. Exiting program.\n");
     exit(-1);
  }
     
  if ( (A = (double*)calloc(tp.numOfBitStreams,sizeof(double))) == NULL ) {
     printf("Final Analysis Report aborted due to insufficient workspace\n");
     fflush(stdout);
     exit(-1);
  }

  /* Compute Metric 1: Proportion of Passing Sequences */

  count = 0; 		
  sampleSize = tp.numOfBitStreams;

  if (test == 12 || test == 13) { /* Special Case: Random Excursion Tests */
     T = (double*)calloc(tp.numOfBitStreams,sizeof(double));
     if (T == NULL) {
        printf("Final Analysis Report aborted due to insufficient workspace\n");
        fflush(stdout);
        exit(-1);
     }
     for(j = 0; j < sampleSize; j++) {
        if(fscanf(fp,"%f",&c)==EOF) {
           printf("Unexpected EOF while reading excursions data\n");
           exit(-1);
        }

        if ((double)c > -0.000001) {
           T[count] = c;
           count++;
        }   
     }
     if(count != 0) {
           free(A);
           if((count!=0) && (A = (double*)calloc(count,sizeof(double)))==NULL) {
             printf("Final Analysis Report aborted due to insufficient workspace\n");
             fflush(stdout);
             exit(-1);
           }
     }

     for(j = 0; j < count; j++)
 	A[j] = T[j];

     sampleSize = count;
     count = 0;

     for(j = 0; j < sampleSize; j++)
        if (A[j] < ALPHA) count++;

     free(T);
  }
  else {
     for(j = 0; j < sampleSize; j++) {
        if(fscanf(fp,"%f",&c) == EOF) {
          printf("Unexpected EOF while reading test data\n");
          exit(-1);
        }
        if (c < ALPHA) count++;
        A[j] = c;
     }
  }
  proportion = 1.0-(double)count/sampleSize;
  passRateMin = 0.99-3.0*sqrt(0.01*(1.0-ALPHA)/((double) sampleSize));
  passRateMax = 0.99+3.0*sqrt(0.01*(1.0-ALPHA)/((double) sampleSize));

  /* Compute Metric 2: Histogram */

  qsort((void*)A,sampleSize,sizeof(double),cmp);
  for(j = 0; j < sampleSize; j++) {
     pos = (int)floor(A[j]*10);
     if (pos == 10) pos--;
     freqPerBin[pos]++;
  }

  chi2 = 0.0;

  if((test == 15) && tp.useCorrectedLZ) { /*Special case for Lempel-Ziv, as per KUH */
     for(j = 0; j < 10; j++) {
       expCount = sampleSize*LZ_S[j];
       chi2 += pow(freqPerBin[j]-expCount,2)/expCount;
     }
  }
  else {
     expCount = sampleSize/10;
     for(j = 0; j < 10; j++)
       chi2 += pow(freqPerBin[j]-expCount,2)/expCount;
  }

  uniformity = igamc(9.0/2.0,chi2/2.0);

  for(j = 0; j < 10; j++)			/* DISPLAY RESULTS */
     fprintf(post,"%3d ", freqPerBin[j]);

  if (uniformity < 0.0001)
     fprintf(post," %f * ", uniformity);
  else
     fprintf(post," %f   ", uniformity);

  if ((proportion < passRateMin) || (proportion > passRateMax))
     fprintf(post,"%6.4f *  %s", proportion, testNames[test]);
  else
     fprintf(post,"%6.4f    %s", proportion, testNames[test]);

  /* Flag modified versions on final report*/
  if((test == 15) && tp.useCorrectedLZ)
     fprintf(post, " (KUH)\n");
  else
     fprintf(post, "\n");
   
  fflush(post);
  fclose(fp);
  free(A);
  return sampleSize;
}

