#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "../include/fips186.h"
#include "../include/config.h"
#include "../include/cephes.h"
#include "../include/defs.h"

/* This is a mostly rewritten approximate entropy test that has a several 
 * advantages over the stock version:
 * It does not require that the entire random string be in memory at once.
 * This is useful in instances where n is very large (e.g. 1,000,000,000 bits)
 *
 * It only goes through each random string once, calculating all rounds of
 * phi at once.  Explanation by example: this is possible because if the 
 * current block (curBlock) is for (m+1) bits, the current block for (m) 
 * bits is just curBlock>>1.  Now imagine you have a table of counts of
 * the number of instances of each pattern of (m+1) sized blocks.  The
 * count of instances of (m) sized blocks is just all of the block sets
 * for the same high order bit pattern.  As there is only a one bit 
 * difference between blocks of (m+1) and blocks of (m), we can just treat
 * table as pairs of counts that must be added for the 'm' block size.
 * This is generalizable all the way to bits of 1 block, where only the
 * MSB of the string matters, so the first half of the table is added to
 * get the number of '0' blocks, and the second half is added to get the
 * total number of '1' blocks.  This allows us to only count the bitstring
 * once (with the highest block size necessary for the test) and then to
 * extract the other numbers nearly "for free".  (They don't require looking
 * at the very large bit stream again.
 *
 * It makes use of a few test specific characteristics that make it so that
 * each block (for both block sizes) is much quicker to compute. 
 * The next large block block is just a shifted current large block
 * with a new LSB, and with the MSB trimmed off.
 * The smaller block size count is derived from the larger block size count.
 *
 * It does all the rounds using the same setup, so there isn't setup and
 * tare down for each round.
 *
 * This version works on raw binary, rather than the epsilon
 * used with most of the sts package.  This helps avoid the
 * computationally expensive steps that accomplished the binary to 
 * array conversion, and the array to binary operations.
 *
 * This version is hard-coded to use the FIPS 186-2 RNG, which is useful
 * principally to test apen parameter selection, rather than useful
 * RNG testing.  It would be fairly easy to modify the code to use other
 * PRNGs, though it wouldn't be very useful to load data off the disk...
 * (the original sts apen test would likely be just as good in this case!)
 *
 * This code is from 30 to over 200 times faster than sts (just running
 * the apen test), so it is very suited for parameter testing.
 */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
                A P P R O X I M A T E  E N T R O P Y   T E S T
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
inline void populateP(unsigned int *P2, int n, int m);
inline void getNewData(UINT32 *X);
void jApEn(unsigned int largestBlockSize, unsigned int n, unsigned int rounds, 
           FILE ** stats, FILE ** results, FILE * checkpoints);
void ahtopb (char *ascii_hex, BYTE *p_binary, int bin_len);
void printnum(FILE *fp, char *S, BYTE *A, int L);

/*Global variables*/
BYTE xkey[64];
unsigned int b;

int main(int argc, char *argv[])
{
    int n, rounds, cntr, largeBlockSize;
    char statsFileName[256], resultsFileName[256], checkpointsFileName[256];
    FILE *stats[31], *results[31], *checkpoints;

    /*Deal with the command line options*/
    if(argc != 6) {
        printf(
         "japen <XKEY> <n> <rounds> <largest APEN block size> <exp dir>\n");
        exit(1);
    }

    b = strlen(argv[1])/2;

    if((b<20) || (b>64)) {
        fprintf(stderr, "Invalid xkey size\n");
        exit(1);
    }

    if(strlen(argv[1]) != b*2) {
        fprintf(stderr, "xkey not in valid hex format\n");
        exit(1);
    }

    ahtopb(argv[1], xkey, b);

    n = atoi(argv[2]);
    rounds = atoi(argv[3]);
    largeBlockSize = atoi(argv[4]);

    /*some quick checks*/
    assert(n>0);
    assert(largeBlockSize >= 0);
    assert(rounds>0);

    assert(n>largeBlockSize);
    assert(largeBlockSize<=30);

    /*setup the directory, if necessary*/
    sprintf(statsFileName, "experiments/%s", argv[5]);
    mkdir(statsFileName, 0777);

    /*Open all the data files*/
    for(cntr=0; cntr <= largeBlockSize; cntr++) {
        sprintf(statsFileName, "experiments/%s/apen-%d", argv[5], cntr);
        mkdir(statsFileName, 0777);

        sprintf(statsFileName, "experiments/%s/apen-%d/stats", argv[5], cntr);
        sprintf(resultsFileName, "experiments/%s/apen-%d/results", 
         argv[5], cntr);

        if((stats[cntr]=fopen(statsFileName, "w"))==NULL) {
            printf("Can't open %s for write\n", statsFileName);
            exit(1);
        }

        if((results[cntr]=fopen(resultsFileName, "w"))==NULL) {
            printf("Can't open %s for write\n", resultsFileName);
            exit(1);
        }
    }

    sprintf(checkpointsFileName, "experiments/%s/checkpoints", argv[5]);
    if((checkpoints=fopen(checkpointsFileName, "w"))==NULL) {
        printf("Can't open %s for write\n", checkpointsFileName);
        exit(1);
    }

    /*Run the tests*/
    jApEn(largeBlockSize, n, rounds, stats, results, checkpoints);

    /*Clean up*/
    for(cntr=0; cntr<=largeBlockSize; cntr++) {
        fclose(results[cntr]);
        fclose(stats[cntr]);
    }

    fclose(checkpoints);

    return(0);
}

/*This is the function that actually calculates the approximate entropy*/
/* Takes :
 * largestBlockSize: the largest block size that is going to be 
 * computed (the 'm' in ApEn(m))
 * n: the number of bits per bitstream
 * rounds: the number of rounds to compute (one result per round)
 * stats: an array of FILE pointers for the "stats" files
 * results: an array of FILE pointers for the "results" files
 * checkpoints: a FILE pointers for the "checkpoints" files
 * 
 */
void jApEn(unsigned int largestBlockSize, unsigned int n, unsigned int rounds, 
           FILE ** stats, FILE ** results, FILE * checkpoints)
{
    double numOfBlocks, chi_squared, p_value;
    double apen, phi[32];
    double log2, logn;

    unsigned int remaining[32], initialRemaining[32], total[32];
    unsigned int countedBlocks[32];

    unsigned int *curRem, *curIR, *curTotal, *curCB;
    double *curPhi, *lastPhiEntry;

    unsigned int *P2, *curPos2, curRound;

    unsigned int powLen2;
    unsigned int *lastTableEntry2;
    int cntr;

    /*All sorts of bit madness here, written for a 32 bit int*/
    assert(sizeof(UINT32) == 4);
    /*If m>=(sizeof(UINT32)*8-2), powLen2 will overflow */
    assert(largestBlockSize <= 30);
    assert(n >= (largestBlockSize+1));

    /*check for silly parameters*/
    assert(rounds > 0);
    assert(n > 0);

    for(cntr=0; cntr<=largestBlockSize; cntr++) {
        assert(stats[cntr]);
        assert(results[cntr]);
    }

    assert(checkpoints);

    /*powlen2 is the number of possible (largestBlockSize+1) 
     * bit blocks.  This is just a quick way of accomplishing:
     powLen2 = pow(2, largestBlockSize+1)*/    
    powLen2 = 1 << (largestBlockSize+1);

    /*Setup the memory*/
    /*P2 is for the (largestBlockSize+1) bit patterns*/
    if ((P2 = (unsigned int *) malloc(powLen2*sizeof(unsigned int))) == NULL) {
        fprintf(stats[0], "ApEn:  Insufficient memory available.\n");
        fflush(stats[0]);
        return;
    }

    /*Save the address of the last entry of P2*/
    lastTableEntry2 = P2 + powLen2 - 1;

    /*phi[0] is zero by definition*/
    phi[0] = 0.0;

    /*Initialize the initialRemaining array; this is the number of P2 table
     * entries that will make up one "count" for this block size.*/
    for(cntr=0; cntr<=largestBlockSize+1; cntr++) {
        initialRemaining[cntr] = 1 << ((largestBlockSize+1) - cntr);
    }

    /*We'll never count any blocks of 0 bits...*/
    countedBlocks[0] = 0;

    /*Some initial constants for speed purposes*/
    numOfBlocks = (double)n;
    log2 = log(2);
    logn = log(numOfBlocks);

    lastPhiEntry = phi + largestBlockSize + 1;

    /*Do all the rounds with the same memory*/
    for (curRound = 0; curRound < rounds; curRound++) {
        /*To facilitate restarts...*/
        printnum(checkpoints, "", xkey, b);
        fflush(checkpoints);

        /*Get the count of the largest block size*/
        /*All other counts are derived from this quantity*/
        populateP(P2, n, largestBlockSize+1);

        /*Calculate this round's ApEn*/

        /*First, initialize the various book-keeping variables*/
        for(cntr=1; cntr<=largestBlockSize+1; cntr++) {
            phi[cntr] = 0.0;
            remaining[cntr] = initialRemaining[cntr];
            total[cntr] = 0;
            countedBlocks[cntr] = 0;
        }

        /*Now get the counts using a non-indexed method*/
        /*This is nearly a wash with a small number of rounds on a
         *Intel P6 based platform (due to register starvation issues, 
         *likely) but it becomes a definite 'win' as the number of 
         *rounds increases*/
        for (curPos2 = P2; curPos2 <= lastTableEntry2; curPos2++) {
            /* First init all the bookkeeping variables for calculating the 
             * phi values*/
            /* curTotal is a pointer to the current location in the total
             * array.  The total array keeps track of how many blocks have
             * been counted for the active logical count of this sized blocks
             */
            curTotal = total+1;

            /* curCB is a pointer to the current location in the countedBlocks
             * array.  The total array keeps track of how many blocks have
             * been counted, total.  The expected number is 'n' at the end*/
            curCB = countedBlocks+1;

            /* curRem is a pointer to the current location in the remaining
             * array.  The remaining array keeps track of how many additional
             * table locations need to be added together*/
            curRem = remaining+1;


            /* curIR is a pointer to the current location in the 
             * initialRemaining array.  The initialRemaining array keeps 
             * track of how many table locations should to be added together
             * for each block size.*/
            curIR = initialRemaining+1;

            /*Now, update each phi value, and keep bookkeeping vars up to date*/
            for(curPhi=phi+1; curPhi<=lastPhiEntry; curPhi++) {
                (*curTotal) += (*curPos2);
                (*curCB) += (*curPos2);
                (*curRem) --;

                if(*curRem == 0) {
                    /*This is the last block; re-init curRem 
                     * If we total shows blocks, add them into the phi 
                     * array and re-init total*/
                    *curRem = *curIR;
                    if(*curTotal != 0) {
                        (*curPhi) += ((double) (*curTotal)) * log(*curTotal);
                        (*curTotal) = 0;
                    }
                }

                /*Keep all the various counters up to date*/
                curTotal++;
                curCB++;
                curRem++;
                curIR++;
            }
        }

        /*We're done making the table, and counting it WRT each PHI value*/
        /*Now, do the final calculations*/
        for(cntr=1; cntr<=largestBlockSize+1; cntr++) {
            /*finish up calculations to yield phi*/
            phi[cntr] = (phi[cntr] / numOfBlocks) - logn;

            /*Calculate the apen*/
            apen = phi[cntr-1] - phi[cntr];

            /*Get p-value from the ApEn value*/
            chi_squared = 2.0 * numOfBlocks * (log2 - apen);
            p_value = igamc(pow(2.0, cntr - 2), chi_squared / 2.0);

            /*output the results*/
            fprintf(stats[cntr-1], "\t\t\tAPPROXIMATE ENTROPY TEST\n");
            fprintf(stats[cntr-1],
                    "\t\t--------------------------------------------\n");

            fprintf(stats[cntr-1], "\t\tCOMPUTATIONAL INFORMATION:\n");
            fprintf(stats[cntr-1],
                    "\t\t--------------------------------------------\n");
            fprintf(stats[cntr-1], "\t\t(a) m (block length)    = %d\n", cntr-1);

            fprintf(stats[cntr-1], "\t\t(b) n (sequence length) = %d\n", n);
            fprintf(stats[cntr-1], "\t\t(b.1) Counted patterns: 1: %u, 2: %u\n", countedBlocks[cntr-1], countedBlocks[cntr]);
            fprintf(stats[cntr-1], "\t\t(b.2) Patterns per bin: 1: %u, 2: %u\n", initialRemaining[cntr-1], initialRemaining[cntr]);
         
            fprintf(stats[cntr-1], "\t\t(c) Chi^2               = %f\n", chi_squared);
            fprintf(stats[cntr-1], "\t\t(d) Phi(m)	       = %f\n", phi[cntr-1]);
            fprintf(stats[cntr-1], "\t\t(d.1) Phi(m)	       = %f bits\n", -phi[cntr-1]/log2);
            fprintf(stats[cntr-1], "\t\t(e) Phi(m+1)	       = %f\n", phi[cntr]);
            fprintf(stats[cntr-1], "\t\t(e.1) Phi(m+1)	       = %f bits\n", -phi[cntr]/log2);
            fprintf(stats[cntr-1], "\t\t(f) ApEn                = %f\n", apen);
            fprintf(stats[cntr-1], "\t\t(f.1) ApEn                = %f bits\n", apen/log2);
            fprintf(stats[cntr-1], "\t\t(g) Log(2)              = %f\n", log2);
            fprintf(stats[cntr-1],
                    "\t\t--------------------------------------------\n");
            fprintf(stats[cntr-1], "%s\t\tp_value = %f\n\n", (p_value < ALPHA)?"FAILURE":"SUCCESS", p_value);
            fprintf(results[cntr-1], "%f\n", p_value);

            /*Just to be safe...*/
            fflush(stats[cntr-1]);
            fflush(results[cntr-1]);
        }
    }

    /*Clean up at the end, when all the rounds are done*/
    free(P2);
    return;
}

/* Count number of instances of each bit m-bit pattern*/
/* This does the heavy lifting, and is binary string intensive
 */
inline void populateP(unsigned int *P2, int n, int m)
{
    UINT32 X[5];
    UINT32 blockBitMask;
    UINT32 firstBlock;
    UINT32 *inputBlock;
    unsigned int inputBitPos;
    UINT32 inputBitMask;
    UINT32 block2;
    unsigned int *curPos2;
    unsigned int *lastTableEntry2;
    unsigned int *lastInputBlock;
    unsigned int i;

    /*Used for clearing out the P2 array*/
    lastTableEntry2 = P2 + ((1 << m) - 1);

    /*Note that P2 has a non-zero, even number of members*/
    /*Clear out P2.*/
    for (curPos2=P2; curPos2<lastTableEntry2; curPos2++) {
        *curPos2 = 0;
        curPos2++;
        *curPos2 = 0;
    }

    /*Use this mask to grab the "active" bits for the larger block (of m bits)
      This should be all 1s for the least significant m bits*/
    blockBitMask = (1 << m) - 1;

    /*Start looking at bitstream*/
    /*Processing occurs in three steps: Beginning, middle and end. */
    /* Beginning processing. Accomplish the following tasks once:
     * 1. Get the first block from the RNG.  This is assumed to fill all of X.
     * X is a series of UINT32s in local endianness (we will be doing bit 
     * operations on the members of X).  Save the first block for the end 
     * processing.  Get a pointer to the last block in the array for later 
     * comparison.
     * 2. Initialize inputBitPos.  This is an integer that refers to a bit 
     * position the inputBlock.  Generally, this refers to the last bit that 
     * was used, though there is one point where it is used to indicate 
     * which bit is ready to use.  (this change is flagged in comments).  
     * Bit numbering is considered to be LSB (bit 0) to MSB (bit 31).
     * 3. Initialize inputBitMask.  This is a bit mask with only the 
     * inputBitPos bit set to 1.
     * 4. Initialize block2.  block2 is the current m bit block.  
     * It is initialized to the first (most significant) m bits in the 
     * inputBlock.
     */

    /*Grab data*/
    getNewData(X);
    inputBlock = X;
    /*This is the address of the last block of data*/
    lastInputBlock = X+4;

    /*keep the first block around for wrapping*/
    firstBlock = *inputBlock;

    /*For the first block, do the initialization*/
    inputBitPos = 31 - (m-1);
    inputBitMask = 0x80000000 >> (m-1);

    /*Only want the most significant m bits*/
    block2 = ((*inputBlock) >> inputBitPos) & blockBitMask;

    /*printf("found 1:%u 2:%u\n", block2>>1, block2);*/

    P2[block2]++;

    /* Middle processing is going to be the majority of processing*/
    /* All blocks other than the first block (which requires initialization),
     * and the last 'm-1' blocks (which require returning to the first 'm-1'
     * bits in the pattern) are processed in this stage.
     * This stage involves identifying the bit that is to be
     * shifted into block2, then updating block2, and then recording
     * block2.
    */
    for (i = 1; i < n-(m-1); i++) {
        /*Make inputBitMask, inputBitPos, and inputBlock point to
         *the bit that should be used, rather than the last used bit*/
        /*Update inputBitMask and inputBitPos.  may update inputBlock.*/
        if(inputBitPos != 0) {
            /*Staying within the same inputBlock*/
            inputBitPos--;
            inputBitMask = inputBitMask >> 1;
        } else {
            /*Done with the current inputBlock.  Refresh it.*/
            if(inputBlock < lastInputBlock) {
                /*Not out of blocks yet.  Just use the next in the X array*/
                inputBlock++;
            } else {
                /*out of blocks. Get a new set from the RNG.  This updates X*/
                getNewData(X);
                /*Point to the first block in X*/
                inputBlock = X;
            }

            /*We have a new block.  Reset bookkeeping data*/
            inputBitPos = 31;
            inputBitMask = 0x80000000;
        }
        /*inputBitMask, inputBitPos, and inputBlock now point to
         *the bit that should be shifted into the LSB of block2*/

        /*update block2*/
        /*shift over to allow for a new LSB*/
        block2 = block2 << 1;

        /*bring in the new LSB*/
        block2 |= (*inputBlock & inputBitMask) >> inputBitPos;

        /*We have now used the bit pointed to by inputBitMask, 
         *inputBitPos, and inputBlock */

        /*Keep only the active bits*/
        block2 &= blockBitMask;

        /*Update P2*/
        /*printf("found 1:%u 2:%u\n", block2>>1, block2);*/
        P2[block2]++;
    }


    /*The end processing takes bits from the first block of data
     * (it makes the bitstring circular).  Only the first block
     * will be necessary, as the blocksize is 30 or less 
     */

    /*Start with the first (most significant) bit of the first block*/
    inputBitPos = 31;
    inputBitMask = 0x80000000;

    for(i=n-(m-1); i < n; i++) {

        /*update block2*/
        /*shift over to allow for a new LSB*/
        block2 = block2 << 1;

        /*bring in the new LSB*/
        block2 |= (firstBlock & inputBitMask) >> inputBitPos;

        /*Keep only the active bits*/
        block2 &= blockBitMask;
  
        /*printf("found 1:%u 2:%u\n", block2>>1, block2);*/
        /*Update P2 */
        P2[block2]++;

        /*We have to stay within the last firstBlock*/
        inputBitPos--;
        inputBitMask = inputBitMask >> 1;
    }

    return;
}

/* Quick function that grabs the data from the FIPS 186-2 PRNG
 * and converts each 4 byte chunk into an integer (in local endianess
 */
inline void getNewData(UINT32 *X)
{
	BYTE rawx[20];

	FIPS186_31(xkey, NULL, b, rawx, SHA_G);
	/*printnum(stdout, "X from RNG: ", rawx, 20);*/

	X[0] = FIPS186_betoh(((UINT32 *)rawx)[0]);
	X[1] = FIPS186_betoh(((UINT32 *)rawx)[1]);
	X[2] = FIPS186_betoh(((UINT32 *)rawx)[2]);
	X[3] = FIPS186_betoh(((UINT32 *)rawx)[3]);
	X[4] = FIPS186_betoh(((UINT32 *)rawx)[4]);
}

void ahtopb (char *ascii_hex, BYTE *p_binary, int bin_len)
{
	BYTE    wbyte, wbyte2;
	int     i; 
	
	for (i = 0; i < bin_len; i++) 
	{
        wbyte = ascii_hex[i * 2];
	    if (wbyte > 'F')
	        wbyte -= 0x20;   
	    if (wbyte > '9')
	        wbyte -= 7;      
	    wbyte -= '0';   
	    wbyte2 = wbyte << 4;

	    wbyte = ascii_hex[i * 2 + 1];
	    if (wbyte > 'F')
			wbyte -= 0x20;
        if (wbyte > '9')
            wbyte -= 7;   
        wbyte -= '0';
		wbyte2 += wbyte;

        p_binary[i] = wbyte2; 
	}
}

void printnum(FILE *fp, char *S, BYTE *A, int L)
{
 int i;

 fprintf(fp, "%s",S);

  for (i=0;i < L/(signed int)sizeof(BYTE); i++) {
      fprintf(fp, "%02x",A[i]);
  }
  fprintf(fp, "\n");

}

void printbits(char *s, unsigned int a) {
	unsigned int mask = 0x80000000;
	int cntr;

	printf("%s", s);

	for(cntr=1; cntr <= 32; cntr++) {
		printf("%s", (a&mask)?"1":"0");
		if(!(cntr%4)) printf(" ");
		mask = mask >> 1;
	}

	printf("\n");
}
