/*
 * file: mp.c
 *
 *
 * DESCRIPTION
 *
 * These functions comprise a multi-precision integer arithmetic
 * and discrete function package.
 *
 *
 * INTEGER FORMAT 
 *
 * Multi-precision integers are represented using arrays of digits,
 * where a digit is the user defined type DIGIT, arranged in a High-low
 * ordering where the most significant digit is stored in the
 * first (lowest addressed) element of the array and the least significant
 * digit is stored in the last (highest addressed) element of the array.
 * **NOTE that the ordering of the bytes within a DIGIT is architecture
 * dependent.
 *
 * The routines given here make the assumption that a DIGIT is a 16-bit
 * word (2 bytes), which on most C compilers is the type "unsigned short".
 * This facilitates word X word multiplication which yields a 32-bit word
 * result (4 bytes), which on most C compilers is the type "unsigned long".
 * This library defines the type DBLWORD to represent 32-bit words.
 *
 * LIMITATIONS
 *
 * Currently few C compilers support a 64-bit integral type which is 
 * unfortunate since many machine architectures have hardware support for
 * performing 32-bit X 32-bit multiplication.  Thus, in order to remain as
 * portable as possible within the  bounds of the C language, this library
 * is limited to 16-bit X 16-bit multiplication which is not optimal for
 * most modern architectures.
 * 
 *
 *
 */


#include 	<stdio.h>
#include	"../include/config.h"
#include	"../include/mp.h"


/*****************************************
** bzero - Fill array with zero's        *
**                                       *
** Parameters:                           *
**                                       *
**  x      Address of array x            *
**  l      Length of array in bytes      *
**                                       *
******************************************/ 
/*
void
bzero(x,l)
  DIGIT x[];
  int l;
{
  l /= sizeof(DIGIT);
  while (l--) x[l] = 0;
}
*/


/*****************************************
** bcopy - Copy one array into another   *
**                                       *
** Parameters:                           *
**                                       *
**  x      Address of array x            *
**  y      Address of array y            *
**  l      Amount of bytes to copy       *
**                                       *
******************************************/ 
/*
void
bcopy(x,y,l)
  DIGIT x[], y[];
  int l;
{
  l /= sizeof(DIGIT);
  while (l--) y[l] = x[l];
}
*/


/*****************************************
** less - Test if x < y                  *
**                                       *
** Returns TRUE (1) if x less than y,    *
** otherwise FALSE (0).                  *
**                                       *
** Parameters:                           *
**                                       *
**  x      Address of array x            *
**  y      Address of array y            *
**  l      Length both x and y in bytes  *
**                                       *
******************************************/ 
int
less(x,y,l)
  DIGIT *x, *y;
  int l;
{
  int i;

  l /= sizeof(DIGIT);
  for (i=0;i<l && x[i] == y[i];i++);
  if (i==l)
    return(0);
  if (x[i] < y[i])
    return(1);
  return(0);
}


/*****************************************
** bshl - shifts array left              *
**                  by one bit.          *
**                                       *	
** x = x * 2                             *
**                                       *
** Parameters:                           *	
**                                       *
**  x      Address of array x            *
**  l      Length array x in bytes       *
**                                       *
******************************************/ 

DIGIT
bshl(x,l)	
  DIGIT *x;
  int l;
{
  DIGIT *p;
 int c1,c2;

  p = x + l/sizeof(DIGIT)-1;
  c1 = 0;
  c2 = 0;
  while (p != x)
  {
    if (*p & (DIGIT)MSBITNDIGIT)
       c2 = 1;
    *p <<= 1;  /* shift the word left once (ls bit = 0) */
    if (c1)
      *p |= 1;
    c1 = c2;
    c2 = 0;
    p--;
  }

  if (*p & (DIGIT)MSBITNDIGIT)
       c2 = 1;
  *p <<= 1;  /* shift the word left once (ls bit = 0) */
   if (c1)
     *p |= (DIGIT)1;

  return (DIGIT)c2;
}



/*****************************************
** Mult - Multiply two integers          *
**                                       *
** A = B * C                             *
**                                       *
** Parameters:                           *	
**                                       *
**  A      Address of the result         *
**  B      Address of the multiplier     *
**  C      Address of the multiplicand   *
**  LB      Length of B in bytes         *
**  LC      Length of C in bytes         *
**                                       *
**  NOTE:  A MUST be LB+LC in length     *
**  NOTE:  LB and LC must be even        *
**                                       *
******************************************/
int
Mult(A,B,LB,C,LC)
 DIGIT A[], B[], C[];
 int LB,LC;
{
   unsigned int i,j,k=0,l,m,n,lC, lB;
   DBLWORD  result;
   DIGIT  *tC, *tB;  

tB = B; 
tC = C; 
LB /= sizeof(DIGIT);
LC /= sizeof(DIGIT);

/* point to the MSB of B and C. */
while ( (tB != (B + LB)) && (*tB == 0))
  tB++;
lB = ((B + LB) - tB);
while ( (tC != (C + LC)) && (*tC == 0))
  tC++;
lC = ((C + LC) - tC);
if (!(lB && lC))   /* that is, one of B or C is zero */
   return(2);

l = (LB+LC) - 1;  
m = LB - 1;
n = LC - 1;

  for (i=0; i<lB; i++)
  {
    result = 0;
    for (j=0;j<lC;j++)
    {
      k = l-i-j;
      result = (DBLWORD)A[k] + ((DBLWORD)B[m-i] * C[n-j]) + (result >> NUMLEN);
      A[k] = (unsigned short)result;
    }
    A[--k] = (unsigned short)result >> NUMLEN;
  }
  return (0);
}



void
ModSqr(A, B,LB, M,LM)
  BYTE *A, *B, *M;
  int LB,LM;
{
  Square(A, B,LB);
  Mod(A,(2*LB), M,LM);

}



void
ModMult(A, B,LB, C,LC, M,LM)
  BYTE *A, *B, *C, *M;
  int LB,LC,LM;
{
  Mult(A, B,LB, C,LC);
  Mod(A, (LB+LC), M,LM);
}



/*****************************************
** smult - Multiply array by a scalar.   *
**                                       *
** A = b * C                             *
**                                       *
** Parameters:                           *	
**                                       *
**  A      Address of the result         *
**  b      Scalar (1 DIGIT)              *
**  C      Address of the multiplicand   *
**  L      Length of C in bytes          *
**                                       *
**  NOTE:  A MUST be L+1 in length       *
**                                       *
******************************************/

void
smult(A,b,C,L)
 DIGIT A[], b, C[];
 int L;
{

   int i;
   DBLWORD result;

    if (!b) return;
    result = 0;
    for (i=0;i<L;i++)
    {
      result = *A + ((DBLWORD)b * *C) + (result >> NUMLEN);
      *A = (unsigned short)result;
      A--; C--;
    }
    *A = (unsigned short)result >> NUMLEN;
}



/*****************************************
** Square() - Square an integer          *
**                                       *
** A = B^2                               *
**                                       *
** Parameters:                           *	
**                                       *
**  A      Address of the result         *
**  B      Address of the operand        *
**  L      Length of B in bytes          *
**                                       *
**  NOTE:  A MUST be 2*L in length       *
**                                       *
******************************************/

int
Square(A,B,L)
 DIGIT A[], B[];
 int L;
{
   int i,l,m,lgth_B;
   DBLWORD result=0;
   DIGIT *tB;

    L /= sizeof(DIGIT);
    tB = B;
    while ((*tB == 0) && (tB != (B + L)))
      tB++;
    lgth_B = ((B + L) - tB);
    if (!lgth_B) return(1);

    l = 2*L - 1; /* index of last digit in A */
    m = L - 1;

    for (i=0; i<lgth_B-1; i++)
      smult(&A[l-2*i-1], B[m-i], &B[m-i-1], L-i-1);

    bshl(A,2*sizeof(DIGIT)*L);

    for (i=0; i<lgth_B; i++) 
    {
      result += (DBLWORD)B[m-i] * B[m-i];
      result += (DBLWORD)A[l-2*i];
      A[l-2*i] = (DIGIT)result;
      result = result >> NUMLEN;
      result += (DBLWORD)A[l-2*i-1];
      A[l-2*i-1] = (DIGIT)result;
      result = result >> NUMLEN;
    }
    if ((l-2*i) >= 0)
	A[l-2*i] = (DIGIT)result;
    return (0);
}



/*****************************************
** ModExp - Modular Exponentiation       *
**                                       *
** A = B ** C (MOD M)                    *
**                                       *	
** Parameters:                           *	
**                                       *
**  A      Address of result             *
**  B      Address of mantissa           *
**  C      Address of exponent           *
**  M      Address of modulus            *
**  LB     Length of B in bytes          *
**  LC     Length of C in bytes          *
**  LM     Length of M in bytes          *
**                                       *
**  NOTE: The integer B must be less     *
**        than the modulus M.      	 *
**  NOTE: A must be at least 3*LM        *
**        bytes long.  However, the      *
**        result stored in A will be     *
**        only LM bytes long.            *
******************************************/
int
ModExp(A, B,LB, C,LC, M,LM)
  BYTE  *A; 
  BYTE  *B, *M;
  DIGIT *C;
  int    LB, LC, LM; 
{
  DIGIT wmask;
  DIGIT bits;
                                   
bits = LC*8;
wmask = MSBITNDIGIT;


#ifdef LITTLE_ENDIAN
if (sizeof(DIGIT) == 2)
  A[LM-2] = 1;
else 
  A[LM-1] = 1;
/*    *( (DIGIT *)&A+LM/sizeof(DIGIT)-1 ) = (DIGIT)1;*/
#else
A[LM-1] = 1;
#endif


#ifdef DEBUG_MODEXP
printnum("A = ",A,LM);
#endif

    while (sniff_bit(C,wmask) == 0)
    {
      wmask >>= 1;
      bits--;
      if (wmask == 0)
      {
        wmask = MSBITNDIGIT;
        C++;
      }
    }

    while (bits--)
    {
      
      bzero(A+LM, LM*2);

     /* temp = A*A (MOD M) */
     ModSqr(A+LM, A,LM,  M,LM);

     /* A = lower L bytes of temp */
     bcopy(A+LM*2, A,LM);

#ifdef DEBUG_MODEXP
     printf("\nAFTER MODSQUARE()\n");
     printnum("RAM[0,63] = ",&RAM[0], 64);
     printnum("RAM[64,127] = ",&RAM[64], 64);
     printnum("RAM[128,191] = ",&RAM[128], 64);
     printnum("RAM[192,255] = ",&RAM[192], 64);
     printf("\n********************\n\n\n");
#endif

      if (sniff_bit(C,wmask))
       {
        bzero(A+LM,(LM+LB));
        ModMult(A+LM, B,LB, A,LM,  M,LM);       /* temp = B * A (MOD M) */
        bcopy(A+LM+(LM+LB)-LM,A,LM);  /* A = lower LM bytes of temp */

#ifdef DEBUG_MODEXP
     printf("\nAFTER MODMULT()\n");
     printnum("RAM[0,63] = ",&RAM[0], 64);
     printnum("RAM[64,127] = ",&RAM[64], 64);
     printnum("RAM[128,191] = ",&RAM[128], 64);
     printnum("RAM[192,255] = ",&RAM[192], 64);
     printf("\n********************\n\n\n");
#endif

       }
 
      wmask >>= 1;
      if (wmask == 0)
      {
/*        wmask = MSBITNWORD;*/
        wmask = MSBITNDIGIT;
        C++;
      }
    }
    return(1);
}




/******************************************************
 * Mod - Computes an integer modulo another integer
 *
 * x = x (mod n)
 *
 * Modified by Indovina 9/27 to 
 * work on 8 or 16-bit and to 
 * take variable sized arguments.
 *
 * NEEDS MORE OPTIMIZATION BY TC and MI
 *******************************************************/
void
Mod(x,lenx, n,lenn)
 DIGIT *x, *n;
 DIGIT lenx, lenn;
{
	 DIGIT  *tx, *tn;   /* pointers to MSD of x, n. */
         DIGIT  l;
         DBLWORD  tmpl,  overflow;
	double   	val,tval;
	 DIGIT  temp1, temp2, multflag = 0, mult, lgth_x, lgth_n;
	 DBLWORD  temp3, tMSx, tmpx;


tx = x;
tn = n;


lenx /= sizeof(DIGIT);
lenn /= sizeof(DIGIT);

/* point to the MSD of n - assume that n is non-zero. */
while (*tn == 0)
  tn++;


/* compute the non-zero length of n. */
lgth_n = (n + lenn) - tn;


/* point to the MSD of x - it may be that x is zero. */
while ((tx != (x+lenx)) && (*tx == 0))
  tx++;
/* compute the non-zero length of x. */
lgth_x = (x + lenx) - tx;

#ifdef DEBUG_MOD
	printf("\n\n");
	printnum("MODing x = ",tx,lgth_x*sizeof(DIGIT));
	printnum("by n = ",tn, lgth_n*sizeof(DIGIT));
	printf("\n\n");
#endif

#ifdef DEBUG_MOD
printf("EXPWD = %.15f\n",         (double)EXPWD                           );
printf("*tn << NUMLEN = %lx\n",   (DBLWORD)*tn << NUMLEN                     );
printf("*(tn+1) = %lx\n",         (DBLWORD)*(tn+1)                           );
printf("^|^ = %.15f\n",           (double) (((DBLWORD)*tn << NUMLEN) | ((DBLWORD)
                                  *(tn+1))) );
printf("\n\n");
#endif

tval=(double) (((DBLWORD)*tn << NUMLEN) | ((DBLWORD)*(tn+1)));

/*
val = EXPWD/4294967295.0;
*/

/*
val = (double)EXPWD  /            (((DBLWORD)*tn << NUMLEN ) | ((DBLWORD)*(tn+1)));
*/

/*
val = (double)((ULONG)0x00010000) / (double)((ULONG)0xffffffff);
printf("%f / %.15f = ",(double)((ULONG)65536),(double)((ULONG)0xffffffff));
*/

val = (double)EXPWD/tval;

#ifdef DEBUG_MOD
printf("val = %.15f\n",val);
#endif

/* Loop while x > n */
while (
#ifdef DEBUG_MOD
printf("lgth_x = %d  lgth_n = %d\n",
	lgth_x*sizeof(DIGIT),lgth_n*sizeof(DIGIT)),
#endif
	lgth_x > lgth_n   ||   (lgth_x == lgth_n  &&
        !less(tx,tn,lgth_n*sizeof(DIGIT))) )
{


/* if MSD of n is greater than that of x then combine the two MSD's of
   x. Otherwise, just take *tx as the MSD. */
if ((*tx < *tn) || multflag)
{
  tMSx = (DBLWORD)(*tx) * EXPWD + *(tx+1);
  *tx = 0;
  tx++;
}
else tMSx = *tx & MAXDIGIT;
;
/* Compute the multiplier. */
if (!(mult = (DIGIT) (tMSx * val))) mult++;

/* Now subtract 'mult * N' from X. */
l = lgth_n;
temp2 = 0;          /* Initialize the carry for 'mult * tn'. */
while (l > 1) { 
  temp1 = (unsigned short)((DBLWORD)mult * *(tn+l-1) + temp2);
  if (*(tx+l-1) >= temp1)  /* check that x[i] >= mult * n[i] */

   tmpx = tx[l-1];
  else  {                  /* else borrow from x[i-1]. */
    tmpl = l-1;
    tmpx = tx[tmpl] + EXPWD;

  /* borrow from previous digit, checking to see if it's the MSD one or not
     and seeing whether or not it's zero. If it is zero then borrow from 
     the still previous byte. */ 

    do  {  
      if (tmpl == 1) tMSx--;
      else
	tx[tmpl-1]--;
      tmpl--;
    } while ((tx[tmpl] == MAXDIGIT) && (tmpl != 0));
   } /* end else */

/* Compute the carry from the product of mult * a byte of tn. */
  temp2 = (unsigned short)(((DBLWORD)mult * (*(tn+l-1)) + temp2) >> NUMLEN);

/* Now finally do the subtraction of a byte of tx - a byte of mult * tn. */
  tx[l-1] = (unsigned short)tmpx - temp1;
  l--;
} /* end inner while */

/* Now check that we didn't subtract too much. */

overflow = 0;
multflag = 0;
temp3 = (DBLWORD)mult * (*tn) + temp2;   /* this is the value of 'mult * tn' of the
                                   MSD digit. */
if (tMSx < temp3)  
   overflow++;
*tx = (unsigned short)(tMSx - temp3) & MAXDIGIT;

if ( overflow ) {           /* then add one multiple of n back. */
     add(tx, lgth_n*sizeof(DIGIT), tn,lgth_n*sizeof(DIGIT));

#ifdef DEBUG_MOD_OVERFLOW
     printf("\nOVERFLOW IN MOD!!!!\n");
#endif


/* if the multiple had been one then we must combine the 2 MSD's of tx.
   That is, N was actually bigger than the lgth_n MSD's of tx. */

     if (mult == 1) 
       multflag = 1;
     }  /* end if overflow */


/* point to the MSD of x  */
while ((tx != (x+lenx)) && (*tx == 0))
  tx++;
/* recompute the non-zero length of x. */
lgth_x = (x + lenx) - tx;


  }  /* END while */

#ifdef DEBUG_MOD
printnum("result ===  ",tx,lgth_x*sizeof(DIGIT));
#endif

}  /* END MOD */



/*********************************************
 *** NEW INV 1-20-92 ***

INVERSE()

len = maximum length of A and N.
In case of DSA, len = 20.
In case of RSA, len = 32.

The function is written to expect that X is 
located at RAM[0].
**********************************************/



/* NEW DIV 1-20-92 */
/* formerly known as divforinv.
   quot = x/y and  x <-- x mod y. */




/*****************************************
** sub - Subtract two integers           *
**                                       *
** A = A - B                             *
**                                       *
**                                       *
** Parameters:                           *	
**                                       *
**  A      Address of subtrahend integer *
**  B      Address of subtractor integer *
**  L      Length of A and B in bytes    *
**                                       *
**  NOTE: In order to save RAM, B is     *
**        two's complemented twice,      *
**        rather than using a copy of B  *
**                                       *
******************************************/
int     negate();
void
sub(A,B,L)
 DIGIT *A, *B;
 int L;
{
  /* Negate B */
  negate(B,L);

  /* Add negative B to A (to do the subtraction) */
  add(A, L, B,L);

  /* Set B back to it's original value */
  negate(B,L);
}


/*****************************************
** negate - Negate an integer            *
**                                       *
** A = -A                                *
**                                       *
**                                       *
** Parameters:                           *	
**                                       *
**  A      Address of integer to negate  *
**  L      Length of A in bytes          *
**                                       *
******************************************/
int
negate(A,L)
 DIGIT A[];
 int L;
{

  short i,lw, index;
  DBLWORD accum;


  lw = L/sizeof(DIGIT);
  index = lw - 1;	/* LSD of result */

  /* Take one's complement of A */
  for (i=0; i<lw ; i++)
     A[i] = ~(A[i]);

  /* Add one to get two's complement of A */
   accum = 1;
   while (accum  && index >= 0)
   {
      accum += A[index];
      A[index--] = (unsigned short)accum;
      accum = accum >> NUMLEN;
   }
   return accum;
}



/***********************************
 * add()
 *
 * A = A + B
 *
 * LB must be <= LA
 *
 ***********************************/
DBLWORD
/*DIGIT*/
add(A, LA, B, LB)
DIGIT  *A, *B;
int LA;  /* in bytes */
int LB;  /* in bytes */
{
 short i,indexA,indexB,hLA,hLB;
 DBLWORD accum;


hLA = LA/sizeof(DIGIT);
hLB = LB/sizeof(DIGIT);

indexA = hLA - 1; 	/* LSD of result */
indexB = hLB - 1; 	/* LSD of B */
accum = 0;

for (i = 0 ; i < hLB; i++)  {
	accum += A[indexA];
	accum += B[indexB--];
	A[indexA--] = (unsigned short)accum;
	accum = accum >> NUMLEN; 
}
if (LA > LB)
	while (accum  && indexA >= 0)
	{
		accum += A[indexA];
		A[indexA--] = (unsigned short)accum;
		accum = accum >> NUMLEN;
	}
return(accum);
} 



/*****************************************
** xor - xor one array into another      *
**                                       *
** Parameters:                           *
**                                       *
**  x      Address of array x            *
**  y      Address of array y            *
**  l      Amount of bytes to copy       *
**                                       *
******************************************/
void
xor(A, x, y, l)
  DIGIT A[], x[], y[];
  int l;
{
  l /= sizeof(DIGIT);
  while (l--)  A[l] = x[l] ^ y[l];
}

