#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <assert.h>

#include "../include/defs.h"
#include "../include/proto.h"
#include "../include/config.h"
#include "../include/utilities1.h"
#include "../include/utilities2.h"
#include "../include/generators1.h"
#include "../include/generators2.h"
#include "../include/generators4.h"
#include "../include/globals.h"

/* Changed ref from ANSI() (which was the non-compliant 9.17 test 
 * function) to NewANSI() (which is the 9.31 test function)
 * Dealt with a small safety issue with convertToBits such that it would not
 * overload epsilon if called when epsilon was already full.
 * -- JEH 2004-Jul-14
 *
 * Added validateParams() function to verify that tests are
 * meaningful.
 * Found that convertToBits was taking a significant amount of CPU time
 * for PRNG testing.  Changed convertToBits to be more like 
 * readHexDigitsInBinaryFormat function (which is faster) to perform 
 * this function.
 * Added printParams() function to print out test parameters.
 * -- JEH 2004-AUG-14
 *
 * Changed DES() to NewFIPS186_DES() and SHA() to NewFIPS186_SHA(), which
 * are compliant FIPS 186-2 Ap 3.1 (general purpose) RNGs, using the
 * DES and SHA based G functions (respectively).  The previously existing
 * implementations did not pass CAVS testing, whereas these versions do.
 * -- JEH 2004-Sep-30
 *
 * Modified the acceptable range for the approximate entropy test
 * to be consistent with our testing.
 *
 * Changed the routines that interact with the operator to allow for
 * input from the configuration file.  If the user specifies values,
 * they are saved in the config file for reporting purposes.
 * 
 * Made openOutputStreams responsible for opening the test-wide grid,
 * finalAnalysisReport, and stats files.  These files now reside in the
 * test directory, in order to allow multiple invocations of the software
 * to execute in the same directory structure.
 * Also made openOutputStreams make data directories in the case where
 * they didn't exist. (This may lead to a porting problem)
 * 
 * Moved the "Number of BitStreams" question to fixParameters.
 * fixParameters can also (in the case of an incomplete config file)
 * ask for 'n'.  
 *
 * Modified validateParams to reflect apen blocksize findings.
 *
 * Added configuration file parsing.  This parsing is really quite
 * stupid, but it is quick, and better than nothing.
 * -- JEH 2004-Oct-07
 *
 * Added option for post-processing only "generator" to assist in
 * (very) long testing groups
 * -- JEH 2004-Oct-10
 *
 * Make configuration file parsing somewhat more general to allow
 * case sensitive filenames and to prevent errors being output
 * with valid configuration options, and correct option display
 * of fixed up LZ test.
 * -- JEH 2007-Aug-12
 *
 */

/*Private functions for use with processing the config file*/
int isInList(char in, char *list);
void compactString(char *str);

int generatorOptions(char** streamFile)
{
  char file[200];
  int  option = NUMOFGENERATORS+1; 

  while ((option < 0) || (option > NUMOFGENERATORS)) {
     if(config.generatorOption == -1)
        option = displayGeneratorOptions();
     else
        option = config.generatorOption;

     switch(option) {
        case 0 :  if(config.dataFile)
                     *streamFile = config.dataFile;
                  else {
                     printf("\t\tUser Prescribed Input File: ");
                     scanf("%s", file);
                     *streamFile = (char*)calloc(200,sizeof(char));
                     if(*streamFile == NULL) {
                        printf("Couldn't duplicate string: Out of memory\n");
                        fflush(stdout);
                        exit(-1);
                     }
                     sprintf(*streamFile,"%s",file);
                  }
                  break;
        #if (INCLUDE_GENERATORS == 1)
        case 1 :  *streamFile = "FIPS 186 general SHA-1 G";
		  break;
        case 2 :  *streamFile = "Linear-Congruential";
		  break;
        case 3 :  *streamFile = "Blum-Blum-Shub";
		  break;
        case 4 :  *streamFile = "Micali-Schnorr";
		  break;
        case 5 :  *streamFile = "Modular-Exponentiation";
		  break;
        case 6 :  *streamFile = "Quadratic-Congruential-1";
		  break;
        case 7 :  *streamFile = "Quadratic-Congruential-2";
		  break;
        case 8 :  *streamFile = "Cubic-Congruential";
		  break;
        case 9 :  *streamFile = "XOR";  
		  break;
        case 10:  *streamFile = "ANSI-X9.31";
		  break;
        case 11 : *streamFile = "FIPS 186 general DES G";
		  break;

        /* INTRODUCE NEW PRNG NAMES HERE */
        /*
        case 12:  *streamFile = "myNewPRNG";
		  break;
        */

        #endif
        case 12: *streamFile = "Process only";
                 break;
        default:  printf("Error:  Try again!\n");
                  option = NUMOFGENERATORS+1;
                  config.generatorOption = -1;
                  break;
     }
  }

  config.generatorOption = option;

  return option;
}

void invokeTestSuite(int option, char* streamFile)
{
  fprintf(output,"_________________________________________________");
  fprintf(output,"_______________________________\n\n");
  fprintf(output,"\t\tFILE = %s\t\tALPHA = %6.4f", streamFile, ALPHA);
  fprintf(output,"\n________________________________________________");
  fprintf(output,"________________________________\n\n");
  fflush(output);
  if (option != 0) printf("\n\t\tStatistical Testing In Progress.........\n");
  switch(option) {
     case 0 :  fileBasedBitStreams(&streamFile);
	       break;
     #if (INCLUDE_GENERATORS == 1)
     case 1 :  NewFIPS186_SHA();
               break;
     case 2 :  lcg();
	       break;
     case 3 :  bbs();
	       break;
     case 4 :  micali_schnorr();
	       break;
     case 5 :  modExp();
	       break;
     case 6 :  quadRes1();
	       break;
     case 7 :  quadRes2();
	       break;
     case 8 :  cubicRes();
	       break;
     case 9 :  exclusiveOR();
	       break;
     case 10:  NewANSI();
               break;
     case 11:  NewFIPS186_DES();
	       break;
     #endif

     /* INTRODUCE NEW PSEUDO RANDOM NUMBER GENERATORS HERE */

     default:  printf("Error in invokeTestSuite!\n");
	       break;
  }
  printf("\n\t\tStatistical Testing Complete!!!!!!!!!!!!\n");
  return;
}

void openOutputStreams(int option)
{
  int   i;
  char  statsDir[200];
  char  resultsDir[200];
  char  testDir[200];
  char  generatorDir[NUMOFGENERATORS][20] = {"AlgorithmTesting","G-SHA-1", 
        "LCG", "BBS", "MS", "MODEXP", "QCG1", "QCG2","CCG", "XOR", "ANSI-X9.31", 
        "G-DES", "error"};
  char  testNames[NUMOFTESTS+1][50] = {" ", "frequency", "block-frequency", 
	"cumulative-sums", "runs", "longest-run", "rank", "fft", 
        "nonperiodic-templates", "overlapping-templates", "universal", "apen", 
        "random-excursions", "random-excursions-variant", "serial",
        "lempel-ziv", "linear-complexity"};
  int   numOfOpenFiles = 0;
  char  *baseDir;

  if(config.baseDir)
     baseDir = config.baseDir;
  else
     baseDir = generatorDir[option];

  /*First make sure the directory exists...*/ 
  sprintf(testDir, "experiments/%s", baseDir);
  mkdir(testDir, 0777);

  sprintf(testDir, "experiments/%s/finalAnalysisReport", baseDir);
  if((post = fopen(testDir, "w")) == NULL) {
    printf("\t\tCould not open finalAnalysisReport file.\n");
    exit(-1);
  }

  /*Don't need to open all this stuff for write if we're not
   * going to output anything... */

  if(option != 12) {
    /*Open the main stats, finalAnalysisReport and grid files */
    sprintf(testDir, "experiments/%s/stats", baseDir);
    if((output = fopen(testDir, "w")) == NULL) {
      printf("\t\tCould not open stats file.\n");
      exit(-1);
    }

    sprintf(testDir, "experiments/%s/grid", baseDir);
    if((grid = fopen(testDir, "w")) == NULL) {
      printf("\t\tCould not open grid file.\n");
      exit(-1);
    }

    /*Now the files for each test*/
    for(i = 1; i <= NUMOFTESTS; i++) {
       if (testVector[i] == 1) {
          sprintf(statsDir, "experiments/%s/%s/stats", baseDir, testNames[i]);
          sprintf(resultsDir, "experiments/%s/%s/results", baseDir, testNames[i]);

          if (DISPLAY_OUTPUT_CHANNELS) {
             printf("*STATS* DIR => %s\n", statsDir);
         printf("RESULTS DIR => %s\n", resultsDir);
             fflush(stdout);
          }

          /*Make sure the intermediary directory exists*/
          sprintf(testDir, "experiments/%s/%s", baseDir, testNames[i]);
          mkdir(testDir, 0777);

          if ( (stats[i] = fopen(statsDir, "w")) == NULL ) {	/* STATISTICS LOG */
             printf("ERROR: LOG FILES COULD NOT BE OPENED.\n");
             printf("       MAX # OF OPENED FILES HAS BEEN REACHED = %d\n", 
                    numOfOpenFiles);
             printf("-OR-   THE OUTPUT DIRECTORY DOES NOT EXIST.\n");
             fflush(stdout);
             exit(-1);
             } else
             numOfOpenFiles++;
          if ( (results[i] = fopen(resultsDir, "w")) == NULL ) {	/* P_VALUES LOG   */
             printf("ERROR: LOG FILES COULD NOT BE OPENED.\n");
             printf("       MAX # OF OPENED FILES HAS BEEN REACHED = %d\n", 
                    numOfOpenFiles);
             printf("-OR-   THE OUTPUT DIRECTORY DOES NOT EXIST.\n");
             fflush(stdout);
             exit(-1);
         }
      else
             numOfOpenFiles++;
       }
    }
  }
  return;
}

int displayGeneratorOptions()
{
  int option = 0;
   
  #if (INCLUDE_GENERATORS == 1)
  printf("________________________________________");
  printf("________________________________________\n\n");
  printf("\t\t     G E N E R A T O R  O P T I O N S \n");
  printf("________________________________________");
  printf("________________________________________\n\n");
  printf("\t[00] Input File\t\t\t[01] FIPS 186 G Using SHA-1\n");
  printf("\t[02] Linear Congruential\t[03] Blum-Blum-Shub\n");
  printf("\t[04] Micali-Schnorr\t\t[05] Modular Exponentiation\n");
  printf("\t[06] Quadratic Congruential I\t[07] Quadratic Congruential II\n");
  printf("\t[08] Cubic Congruential\t\t[09] XOR\n");
  printf("\t[10] ANSI X9.31 (3-DES)\t\t[11] FIPS 186 G Using DES\n");
  printf("\t[12] Process only (no testing)\n");

  printf("________________________________________");
  printf("________________________________________\n\n");
  printf("\tOPTION ----> "); scanf("%d",&option);
  #endif
  return option;
}

void chooseTests()
{
  int i;

  if(config.testsSelected == 0) {
    displayTests();
    printf("\n\n\n");
    printf("________________________________________");
    printf("________________________________________\n\n");
    printf("\t\t\t     INSTRUCTIONS\n");
    printf("\tEnter 0 if you DO NOT want to apply all of the\n");
    printf("\tstatistical tests to each sequence and 1 if you DO. \n");
    printf("________________________________________");
    printf("________________________________________\n\n");
    printf("\tEnter Choice:  ");
    scanf("%d",&testVector[0]);
    printf("\n");
    if (testVector[0] == 1)
       for(i=1; i <= NUMOFTESTS; i++) testVector[i] = 1;
    else {
       displayTests();
       printf("________________________________________");
       printf("________________________________________\n\n");
       printf("\t\t\t     INSTRUCTIONS\n");
       printf("\tEnter a 0 or 1 to indicate whether or not the numbered\n");
       printf("\tstatistical test should be applied to each sequence.  \n");
       printf("\tFor example, 1111111111111111 applies every test to \n");
       printf("\teach sequence.\n");
       printf("________________________________________");
       printf("________________________________________\n");
       printf("\t\t\t    1234567891111111\n");
       printf("\t\t\t             0123456\n");
       printf("\t\t\t    ");
       for (i = 1; i <= NUMOFTESTS; i++) 
           scanf("%1d", &testVector[i]);
       printf("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n");
    }
  }
  return;
}

void displayTests()
{
   printf("________________________________________");
   printf("________________________________________\n\n");
   printf("\t\t      S T A T I S T I C A L  T E S T S\n");
   printf("________________________________________");
   printf("________________________________________\n\n");
   printf("   [01] Frequency\t\t\t");
   printf("   [02] Block Frequency\n");
   printf("   [03] Cumulative Sums\t\t\t");
   printf("   [04] Runs\n");
   printf("   [05] Longest Run of Ones\t\t");
   printf("   [06] Rank\n");
   printf("   [07] Discrete Fourier Transform\t");
   printf("   [08] Nonperiodic Template Matchings\n");
   printf("   [09] Overlapping Template Matchings\t");
   printf("   [10] Universal Statistical\n");
   printf("   [11] Approximate Entropy\t\t");
   printf("   [12] Random Excursions\n");
   printf("   [13] Random Excursions Variant\t");
   printf("   [14] Serial\n");
   printf("   [15] Lempel-Ziv Complexity\t\t");
   printf("   [16] Linear Complexity\n");
   /*printf("   [17] Overlapping Pairs Sparse Occupancy\n");*/
   return;
}

void fixParameters()
{
   /* Just in case a config file is specified that doesn't have this */
  if(tp.n == -1) {
    printf("\tHow many bits should be in each bitstreams? ");
    scanf("%d", &tp.n);
  }

  if ((testVector[2] == 1) && (tp.blockFrequencyBlockLength == -1)) {
     printf("\tEnter Block Frequency Test block length:  "); 
     scanf("%d",&tp.blockFrequencyBlockLength);
  }
  if ((testVector[8] == 1) && (tp.nonOverlappingTemplateBlockLength == -1)) {
     printf("\tEnter NonOverlapping Template Test Block Length:  "); 
     scanf("%d", &tp.nonOverlappingTemplateBlockLength);
  }
  if ((testVector[9] == 1) && (tp.overlappingTemplateBlockLength == -1)) {
     printf("\tEnter Overlapping Template Test Block Length:  "); 
     scanf("%d", &tp.overlappingTemplateBlockLength);
  }
  if (testVector[10] == 1) {
     if(tp.universalBlockLength == -1) {
       printf("\tEnter Universal Test Block Length:  "); 
       scanf("%d", &tp.universalBlockLength);
     }

     if(tp.universalNumberInitializationSteps == -1) {
       printf("\tEnter Universal Test Number Of Initialization Steps:  "); 
       scanf("%d", &tp.universalNumberInitializationSteps);
     }
  }
  if ((testVector[11] == 1) && (tp.approximateEntropyBlockLength == -1)) {
     printf("\tEnter Approximate Entropy Test Block Length:  "); 
     scanf("%d", &tp.approximateEntropyBlockLength);
  }
  if ((testVector[14] == 1) && (tp.serialBlockLength == -1)) {
     printf("\tEnter Serial Test Block Length:  "); 
     scanf("%d", &tp.serialBlockLength);
  }
  if ((testVector[16] == 1) && (tp.linearComplexitySequenceLength == -1)) {
     printf("\tEnter Linear Complexity Test Subsequence Length:  "); 
     scanf("%d", &tp.linearComplexitySequenceLength);
  }

   /* It wasn't clear why this wasn't done here, so I moved it */
  if(tp.numOfBitStreams == -1) {
    printf("\n\t\tHow many bitstreams should be generated? ");
    scanf("%d", &tp.numOfBitStreams);
  }

  return;
}

/*converted to a byte by byte conversion.  Moved from 32bit int to number 
 * of bytes, which required changing every invocation of this function.
 */
int convertToBits(BYTE* x, int n, int* n0, int* n1, int* bitsRead, int offset, int numBytes)
{
  int      i, c, bit, done=0;
  BYTE displayMask = 1 << 7;
  BYTE buffer;

     assert(sizeof(BYTE) == 1);

     if(*bitsRead >= n) 
       done = 1;

     for(i = 0; (i < numBytes) && !done; i++) {
	buffer=x[i+offset];
        for(c = 1; c <= 8; c++) {
          if (buffer & displayMask) {
            bit = 1;
            (*n1)++;
          } else {
            bit = 0;
            (*n0)++;
          }
          buffer <<= 1;
          epsilon[*bitsRead].b = bit;
          (*bitsRead)++;
          if (*bitsRead == n) {
            done=1;
            break;
	  }
       }
    }
 return done;
}


void validateParams()
{
    int goodParams = 1;

    int MaurerParams[11][2] = {
        {640, 387840},
        {1280, 904960},
        {2560, 2068480},
        {5120, 4654080},
        {10240, 1342400},
        {20480, 22753280},
        {40960, 49643520},
        {81920, 107560960},
        {163840, 231669760},
        {327680, 496435200},
        {655360, 1059061760}
    };

    if (testVector[1] == 1) {
        if (tp.n < 100) {
            printf("Frequency Test requires n >= 100 (SP800-22 sec. 2.1.7)\n");
            goodParams = 0;
        }
    }

    if (testVector[2] == 1) {
        if (tp.n < 100) {
            printf("Frequency Test within a block requires n >= 100\n (SP800-22 sec. 2.2.7)\n");
            goodParams = 0;
        }
        if (tp.blockFrequencyBlockLength < 20) {
            printf
                ("Frequency Test within a block requires block length >= 20\n (SP800-22 sec. 2.2.7)\n");
            goodParams = 0;
        }
        if (tp.blockFrequencyBlockLength <= (0.01 * tp.n)) {
            printf
                ("Frequency Test within a block requires block length > 0.01 * n (> %d)\n (SP800-22 sec. 2.2.7)\n", (int)(0.01 * (float)tp.n));
            goodParams = 0;
        }
    }
    if (testVector[3] == 1) {
        if (tp.n < 100) {
            printf("Cumulative Sums Test requires n >= 100\n (SP800-22 sec. 2.14.7)\n");
            goodParams = 0;
        }
    }
    if (testVector[4] == 1) {
        if (tp.n < 100) {
            printf("Runs Test requires n >= 100 (SP800-22 sec. 2.3.7)\n");
            goodParams = 0;
        }
    }
    if (testVector[5] == 1) {
        if ((LONG_RUNS_CASE_8 == 1) && (tp.n < 128)) {
            printf("Longest Runs Test (CASE_8) requires n>= 128\n (SP800-22 sec. 2.4.2)\n");
            goodParams = 0;
        } else if ((LONG_RUNS_CASE_128 == 1) && (tp.n < 6272)) {
            printf("Longest Runs Test (CASE_128) requires n>= 6272\n (SP800-22 sec. 2.4.2)\n");
            goodParams = 0;
        } else if ((LONG_RUNS_CASE_10000 == 1) && (tp.n < 750000)) {
            printf("Longest Runs Test (CASE_128) requires n>= 750000\n (SP800-22 sec. 2.4.2)\n");
            goodParams = 0;
        }
    }
    if (testVector[6] == 1) {
        if (tp.n < 38912) {
            printf("Binary Matrix Rank Test requires n >= 38912\n (SP800-22 sec. 2.5.7)\n");
            goodParams = 0;
        }
    }
    if (testVector[7] == 1) {
        if (tp.n < 1000) {
            printf
                ("Discrete Fourier Transform (Spectral or FFT) Test requires n >= 1000\n (SP800-22 sec. 2.6.7)\n");
            goodParams = 0;
        }
    }
    if (testVector[8] == 1) {
        if (tp.n != 1000000) {
            printf
                ("Non-Overlapping (non-periodic) Template Matching Test requires n = 1000000\n (SP800-22 sec. 2.7.7)\n");
            goodParams = 0;
        }
        if ((tp.nonOverlappingTemplateBlockLength != 9) &&
            (tp.nonOverlappingTemplateBlockLength != 10)) {
            printf
                ("Non-Overlapping (non-periodic) Template Matching Test requires\n block size of 9 or 10 (SP800-22 sec. 2.7.7)\n");
            goodParams = 0;
        }
    }
    if (testVector[9] == 1) {
        if (tp.n < 1000000) {
            printf
                ("Overlapping Template Matching Test requires n >= 1000000\n (SP800-22 sec. 2.8.7)\n");
            goodParams = 0;
        }
        if ((tp.overlappingTemplateBlockLength != 9) &&
            (tp.overlappingTemplateBlockLength != 10)) {
            printf
                ("Non-Overlapping (non-periodic) Template Matching Test requires\n block size of 9 or 10 (SP800-22 sec. 2.8.7)\n");
            goodParams = 0;
        }
    }
    if (testVector[10] == 1) {
        if ((tp.universalBlockLength < 6)
            || (tp.universalBlockLength > 16)) {
            printf
                ("Maurer's Universal Statistical Test requires 6 <= block size <= 16\n (SP800-22 sec. 2.9.7)\n");
            goodParams = 0;
        } else {
            if (MaurerParams[tp.universalBlockLength - 6][0] !=
                tp.universalNumberInitializationSteps) {
                printf
                    ("Maurer's Universal Statistical Test, L=%d requires Q=%d\n (SP800-22 sec. 2.9.7)\n",
                     tp.universalBlockLength,
                     MaurerParams[tp.universalBlockLength - 6][0]);
                goodParams = 0;
            }
            if (tp.n < MaurerParams[tp.universalBlockLength - 6][1]) {
                printf
                    ("Maurer's Universal Statistical Test, L=%d requires n>=%d\n (SP800-22 sec. 2.9.7)\n",
                     tp.universalBlockLength,
                     MaurerParams[tp.universalBlockLength - 6][1]);
                goodParams = 0;
            }
        }
    }
    if (testVector[11] == 1) {
        if (tp.approximateEntropyBlockLength >
            (((int) floor(log(tp.n) / log(2))) - 11)) {
            printf
                ("Approximate Entropy test requires a block size %d or less for\n for this selection of n (SP800-22 sec. 2.13.7, IGL testing)\n",
                 (((int) floor(log(tp.n) / log(2))) - 11));
            goodParams = 0;
        }
    }
    if (testVector[12] == 1) {
        if (tp.n < 1000000) {
            printf("Random Excursions Test requires n >= 1000000 (SP800-22 sec. 2.15.7)\n");
            goodParams = 0;
        }
    }
    if (testVector[13] == 1) {
        if (tp.n < 1000000) {
            printf
                ("Random Excursions Variant Test requires n >= 1000000\n (SP800-22 sec. 2.16.7)\n");
            goodParams = 0;
        }
    }
    if (testVector[14] == 1) {
        if (tp.serialBlockLength >=
            (((int) floor(log(tp.n) / log(2))) - 2)) {
            printf
                ("Serial test requires a block size less than %d for\n for this\n selection of n (SP800-22 sec. 2.12.7)",
                 (((int) floor(log(tp.n) / log(2))) - 2));
            goodParams = 0;
        }
    }
    if (testVector[15] == 1) {
        if (tp.n != 1000000) {
            printf("Lempel-Ziv Compression Test requires n = 1000000\n (SP800-22 sec. 2.10.7, implementation limitation)\n");
            goodParams = 0;
        }
    }
    if (testVector[16] == 1) {
        if (tp.n < 1000000) {
            printf("Linear Complexity Test requires n >= 1000000 (SP800-22 sec. 2.11.7)\n");
            goodParams = 0;
        }
        if ((tp.linearComplexitySequenceLength < 500)
            || (tp.linearComplexitySequenceLength > 5000)
            || ((tp.n / tp.linearComplexitySequenceLength) < 200)) {
            printf
                ("Linear Complexity Test 500 <= (Sequence Length) <= 5000 and n / (Sequence Length) >= 200 \n (SP800-22 sec. 2.11.7)\n");
            goodParams = 0;
        }
    }

    if (tp.numOfBitStreams < 1000) {
        printf("Warning: number of bit streams < 1000.\n");
        printf
            ("Proportion of Sequences Passing a Test interpretation not valid.\n (SP800-22 sec. 4.2.1)\n");
        goodParams = 0;
    }

    if (goodParams == 1)
        tp.validParams = 1;
    else
        tp.validParams = 0;
}

void printParams(FILE *fp)
{

    int i;

    fprintf(fp, "Parameters are:\n");
    fprintf(fp, "n: %d\n", tp.n);
    fprintf(fp, "numOfBitStreams: %d\n", tp.numOfBitStreams);

    if (testVector[2] == 1) {
        fprintf(fp, "blockFrequencyBlockLength: %d\n",
                tp.blockFrequencyBlockLength);
    }
    if (testVector[8] == 1) {
        fprintf(fp, "nonOverlappingTemplateBlockLength: %d\n",
                tp.nonOverlappingTemplateBlockLength);
    }
    if (testVector[9] == 1) {
        fprintf(fp, "overlappingTemplateBlockLength: %d\n",
                tp.overlappingTemplateBlockLength);
    }
    if (testVector[14] == 1) {
        fprintf(fp, "serialBlockLength: %d\n", tp.serialBlockLength);
    }
    if (testVector[16] == 1) {
        fprintf(fp, "linearComplexitySequenceLength: %d\n",
                tp.linearComplexitySequenceLength);
    }
    if (testVector[11] == 1) {
        fprintf(fp, "approximateEntropyBlockLength: %d\n",
                tp.approximateEntropyBlockLength);
    }
    if (testVector[10] == 1) {
        fprintf(fp, "universalBlockLength: %d\n",
                tp.universalBlockLength);
        fprintf(fp, "universalNumberInitializationSteps: %d\n",
                tp.universalNumberInitializationSteps);
    }
    if(testVector[15] == 1) {
        fprintf(fp, "LZ is %s\n", (tp.useCorrectedLZ==1) ? "corrected" : "original");
    }

    fprintf(fp, "Parameters are: %s\n",
            (tp.validParams == 1) ? "valid" : "invalid");

    fprintf(fp, "Config options as follows:\n");
    if(config.baseDir) fprintf(fp, "Experiment directory: %s\n", config.baseDir);
    if(config.configFile) fprintf(fp, "Config file: %s\n", config.configFile);
    if(config.generatorOption != -1) fprintf(fp, "Generator Option: %d\n", config.generatorOption);
    if(config.dataFile && (config.generatorOption == 0)) fprintf(fp, "Data file: %s\n", config.dataFile);
    if((config.dataFileFormat != -1) && (config.generatorOption == 0)) fprintf(fp, "Data File Format: %d\n", config.dataFileFormat);
    if(config.ansix931.V && (config.generatorOption == 10)) fprintf(fp, "ANSI X9.31 V: %s\n", config.ansix931.V);
    if(config.ansix931.DT && (config.generatorOption == 10)) fprintf(fp, "ANSI X9.31 DT: %s\n", config.ansix931.DT);
    if(config.ansix931.K && (config.generatorOption == 10)) fprintf(fp, "ANSI X9.31 K: %s\n", config.ansix931.K);
    if(config.fips186.xkey && ((config.generatorOption == 11) || (config.generatorOption == 1))) fprintf(fp, "FIPS 186-2 xkey: %s\n", config.fips186.xkey);

    fprintf(fp, "Tests (in test vector format): ");
    for(i=0; i<=NUMOFTESTS; i++) fprintf(fp, "%d", testVector[i]);
    fprintf(fp, "\n");
}

/* Note: strncmp is used when there is a numerical parameter
   strcmp is used when the entire string needs to match
*/
void processConfigFile(const char *configFile) {
    FILE *fp;
    char line[256];
    int i;

    if((fp = fopen(configFile, "r")) == NULL) {
        printf("Can't open config file: %s\n", configFile);
        exit(1);
    }

    while(!feof(fp) && !ferror(fp)) {
        if(fgets(line, sizeof(line), fp)) {
            compactString(line);
            if(line[0]) {
                if(!strncmp("n=", line, 2)) 
                    tp.n=atoi(line+2);
                else if(!strncmp("numofbitstreams=", line, 16)) 
                    tp.numOfBitStreams=atoi(line+16);
                else if(!strcmp("alltests=yes", line)) 
                    for(i=0; i<=NUMOFTESTS; i++) testVector[i] = 1;
                else if(!strcmp("frequency=yes", line))
                    testVector[1] = 1;
                else if(!strcmp("frequency=no", line))
                    testVector[1] = 0;
                else if(!strcmp("blockfrequency=yes", line))
                    testVector[2] = 1;
                else if(!strcmp("blockfrequency=no", line))
                    testVector[2] = 0;
                else if(!strncmp("blockfrequencyblocklength=", line, 26)) 
                    tp.blockFrequencyBlockLength=atoi(line+26);
                else if(!strcmp("runs=yes", line))
                    testVector[4] = 1;
                else if(!strcmp("runs=no", line))
                    testVector[4] = 0;
                else if(!strcmp("longestrunofones=yes", line))
                    testVector[5] = 1;
                else if(!strcmp("longestrunofones=no", line))
                    testVector[5] = 0;
                else if(!strcmp("matrixrank=yes", line))
                    testVector[6] = 1;
                else if(!strcmp("matrixrank=no", line))
                    testVector[6] = 0;
                else if(!strcmp("fft=yes", line))
                    testVector[7] = 1;
                else if(!strcmp("fft=no", line))
                    testVector[7] = 0;
                else if(!strcmp("nonoverlappingtemplatematchings=yes", line))
                    testVector[8] = 1;
                else if(!strcmp("nonoverlappingtemplatematchings=no", line))
                    testVector[8] = 0;
                else if(!strncmp("nonoverlappingtemplateblocklength=", line, 34)) 
                    tp.nonOverlappingTemplateBlockLength=atoi(line+34);
                else if(!strcmp("overlappingtemplatematchings=yes", line))
                    testVector[9] = 1;
                else if(!strcmp("overlappingtemplatematchings=no", line))
                    testVector[9] = 0;
                else if(!strncmp("overlappingtemplateblocklength=", line, 31)) 
                    tp.overlappingTemplateBlockLength=atoi(line+31);
                else if(!strcmp("universal=yes", line))
                    testVector[10] = 1;
                else if(!strcmp("universal=no", line))
                    testVector[10] = 0;
                else if(!strncmp("universalblocklength=", line, 21)) 
                    tp.universalBlockLength=atoi(line+21);
                else if(!strncmp("universalnumberinitializationsteps=", line, 35)) 
                    tp.universalNumberInitializationSteps=atoi(line+35);
                else if(!strcmp("lempelzivcomplexity=yes", line))
                    testVector[15] = 1;
                else if(!strcmp("lempelzivcomplexity=no", line))
                    testVector[15] = 0;
                else if(!strcmp("usecorrectedlz=no", line))
                    tp.useCorrectedLZ=0;
                else if(!strcmp("usecorrectedlz=yes", line))
                    tp.useCorrectedLZ=1;
                else if(!strcmp("linearcomplexity=yes", line))
                    testVector[16] = 1;
                else if(!strcmp("linearcomplexity=no", line))
                    testVector[16] = 0;
                else if(!strncmp("linearcomplexitysequencelength=", line, 31)) 
                    tp.linearComplexitySequenceLength=atoi(line+31);
                else if(!strcmp("serial=yes", line))
                    testVector[14] = 1;
                else if(!strcmp("serial=no", line))
                    testVector[14] = 0;
                else if(!strncmp("serialblocklength=", line, 18)) 
                    tp.serialBlockLength=atoi(line+18);
                else if(!strcmp("approximateentropy=yes", line))
                    testVector[11] = 1;
                else if(!strcmp("approximateentropy=no", line))
                    testVector[11] = 0;
                else if(!strncmp("approximateentropyblocklength=", line, 30)) 
                    tp.approximateEntropyBlockLength=atoi(line+30);
                else if(!strcmp("cusum=yes", line))
                    testVector[3] = 1;
                else if(!strcmp("cusum=no", line))
                    testVector[3] = 0;
                else if(!strcmp("randomexcursions=yes", line))
                    testVector[12] = 1;
                else if(!strcmp("randomexcursions=no", line))
                    testVector[12] = 0;
                else if(!strcmp("randomexcursionsvariant=yes", line))
                    testVector[13] = 1;
                else if(!strcmp("randomexcursionsvariant=no", line))
                    testVector[13] = 0;
                else if(!strncmp("basedir=", line, 8)) 
                    config.baseDir=strdup(line+8);
                else if(!strncmp("generatoroption=", line, 16)) 
                    config.generatorOption=atoi(line+16);
                else if(!strncmp("datafile=", line, 9)) 
                    config.dataFile=strdup(line+9);
                else if(!strcmp("datafileformat=ascii", line))
                    config.dataFileFormat=ASCII_FORMAT;
                else if(!strcmp("datafileformat=binary", line))
                    config.dataFileFormat=BINARY_FORMAT;
                else if(!strncmp("v=", line, 2)) 
                    config.ansix931.V=strdup(line+2);
                else if(!strncmp("dt=", line, 3)) 
                    config.ansix931.DT=strdup(line+3);
                else if(!strncmp("k=", line, 2)) 
                    config.ansix931.K=strdup(line+2);
                else if(!strncmp("xkey=", line, 5)) 
                    config.fips186.xkey=strdup(line+5);
                else if(strcmp("alltests=no", line)) 
                    printf("Unknown config: %s\n", line);
            }
        }
    }

    /*Make sure the testVector is internally consistent*/
    testVector[0]=1;
    config.testsSelected=0;

    for(i=1; i<=NUMOFTESTS; i++) {
        if(testVector[i]==0)
            testVector[0]=0;
        else
            config.testsSelected=1;
    }

    fclose(fp);
}

/*normalize the string: 
 *remove all unexpected chars
 *convert all uppercase letters to lowercase
 *remove all comments
*/
void compactString(char *str) {
    char *curRead = str, *curWrite = str;
    int quotemode=0;

    while(*curRead) {
        if(quotemode && (*curRead != '"')) {
            *curWrite=*curRead;
            curWrite++;
            curRead++;
        } else if(*curRead == '#') {
            *curWrite='\0';
            curRead=curWrite;
        } else if(*curRead == '"') {
            quotemode = quotemode?0:1;
            curRead++;
        } else if(isupper(*curRead)) {
            *curWrite=tolower(*curRead);
            curWrite++;
            curRead++;
        } else if(isInList(*curRead, "=/:-_.") || isalnum(*curRead)) {
            *curWrite=*curRead;
            curWrite++;
            curRead++;
        } else
            curRead++;
    }

    *curWrite = '\0';
}

/*Checks if the provided character is in the provided list*/
int isInList(char in, char *list) {
    while(*list) {
        if(in == *list) return 1;
        list++;
    }

    return 0;
}
